import types
import resource
import psutil
import dask
import numpy as np
import warnings
import gc
import logging
import time
import glob
import os
import subprocess
import sys
import tempfile
from pathlib import Path
from dask import delayed, compute, config
from dask.distributed import Client, LocalCluster
from datetime import datetime as dt, timedelta
from .basic_utils import *


#################################
# Process management
#################################
def get_nprocess_solarpipe(jobid):
    """
    Get numbers of processes currently running

    Parameters
    ----------
    workdir : str
        Work directory name
    jobid : int
        Job ID

    Returns
    -------
    int
        Number of running processes
    """
    cachedir = get_cachedir()
    pid_file = f"{cachedir}/pids/pids_{jobid}.txt"
    pids = np.loadtxt(pid_file, unpack=True)
    n_process = 0
    for pid in pids:
        if psutil.pid_exists(int(pid)):
            n_process += 1
    return n_process


def get_jobid():
    """
    Get Job ID with millisecond-level uniqueness.

    Returns
    -------
    int
        Job ID in the format YYYYMMDDHHMMSSmmm (milliseconds)
    """
    cachedir = get_cachedir()
    jobid_file = os.path.join(cachedir, "jobids.txt")
    if os.path.exists(jobid_file):
        prev_jobids = np.loadtxt(jobid_file, unpack=True, dtype="int64")
        if prev_jobids.size == 0:
            prev_jobids = []
        elif prev_jobids.size == 1:
            prev_jobids = [str(prev_jobids)]
        else:
            prev_jobids = [str(jid) for jid in prev_jobids]
    else:
        prev_jobids = []

    if len(prev_jobids) > 0:
        FORMAT = "%Y%m%d%H%M%S%f"
        CUTOFF = dt.utcnow() - timedelta(days=15)
        filtered_prev_jobids = []
        for job_id in prev_jobids:
            job_time = dt.strptime(job_id.ljust(20, "0"), FORMAT)  # pad if truncated
            if job_time >= CUTOFF or job_id == 0:  # Job ID 0 is always kept
                filtered_prev_jobids.append(job_id)
        prev_jobids = filtered_prev_jobids

    now = dt.utcnow()
    cur_jobid = (
        now.strftime("%Y%m%d%H%M%S") + f"{int(now.microsecond/1000):03d}"
    )  # ms = first 3 digits of microseconds
    prev_jobids.append(cur_jobid)

    job_ids_int = np.array(prev_jobids, dtype=np.int64)
    np.savetxt(jobid_file, job_ids_int, fmt="%d")

    return int(cur_jobid)


def save_main_process_info(pid, jobid, msname, workdir, outdir, cpu_frac, mem_frac):
    """
    Save main processes info

    Parameters
    ----------
    pid : int
        Main job process id
    jobid : int
        Job ID
    msname : str
        Main measurement set
    workdir : str
        Work directory
    outdir : str
        Output directory
    cpu_frac : float
        CPU fraction of the job
    mem_frac : float
        Memory fraction of the job

    Returns
    -------
    str
        Job info file name
    """
    cachedir = get_cachedir()
    prev_main_pids = glob.glob(f"{cachedir}/main_pids_*.txt")
    prev_jobids = [
        str(os.path.basename(i).rstrip(".txt").split("main_pids_")[-1])
        for i in prev_main_pids
    ]
    if len(prev_jobids) > 0:
        FORMAT = "%Y%m%d%H%M%S%f"
        CUTOFF = dt.utcnow() - timedelta(days=15)
        filtered_prev_jobids = []
        for i in range(len(prev_jobids)):
            job_id = prev_jobids[i]
            job_time = dt.strptime(job_id.ljust(20, "0"), FORMAT)  # pad if truncated
            if job_time < CUTOFF or job_id == 0:  # Job ID 0 is always kept
                filtered_prev_jobids.append(job_id)
            else:
                os.system(f"rm -rf {prev_main_pids[i]}")
                if os.path.exists(f"{cachedir}/pids/pids_{job_id}.txt"):
                    os.system(f"rm -rf {cachedir}/pids/pids_{job_id}.txt")
    main_job_file = f"{cachedir}/main_pids_{jobid}.txt"
    main_str = f"{jobid} {pid} {msname} {workdir} {outdir} {cpu_frac} {mem_frac}"
    with open(main_job_file, "w") as f:
        f.write(main_str)
    return main_job_file


def save_pid(pid, pid_file):
    """
    Save PID

    Parameters
    ----------
    pid : int
        Process ID
    pid_file : str
        File to save
    """
    if os.path.exists(pid_file):
        pids = np.loadtxt(pid_file, unpack=True, dtype="int")
        pids = np.append(pids, pid)
    else:
        pids = np.array([int(pid)])
    np.savetxt(pid_file, pids, fmt="%d")


def create_batch_script_nonhpc(cmd, workdir, basename, write_logfile=True):
    """
    Function to make a batch script not non-HPC environment

    Parameters
    ----------
    cmd : str
        Command to run
    workdir : str
        Work directory of the measurement set
    basename : str
        Base name of the batch files
    write_logfile : bool, optional
        Write log file or not

    Returns
    -------
    str
        Batch file name
    str
        Log file name
    """
    batch_file = workdir + "/" + basename + ".batch"
    cmd_batch = workdir + "/" + basename + "_cmd.batch"
    finished_touch_file = workdir + "/.Finished_" + basename
    os.system("rm -rf " + finished_touch_file + "*")
    finished_touch_file_error = finished_touch_file + "_1"
    finished_touch_file_success = finished_touch_file + "_0"
    cmd_file_content = f"{cmd}; exit_code=$?; if [ $exit_code -ne 0 ]; then touch {finished_touch_file_error}; else touch {finished_touch_file_success}; fi"
    if write_logfile:
        if os.path.isdir(workdir + "/logs") == False:
            os.makedirs(workdir + "/logs")
        logfile = workdir + "/logs/" + basename + ".log"
    else:
        logfile = "/dev/null"
    batch_file_content = f"""export PYTHONUNBUFFERED=1\nnohup sh {cmd_batch}> {logfile} 2>&1 &\nsleep 2\n rm -rf {batch_file}\n rm -rf {cmd_batch}"""
    if os.path.exists(cmd_batch):
        os.system("rm -rf " + cmd_batch)
    if os.path.exists(batch_file):
        os.system("rm -rf " + batch_file)
    with open(cmd_batch, "w") as cmd_batch_file:
        cmd_batch_file.write(cmd_file_content)
    with open(batch_file, "w") as b_file:
        b_file.write(batch_file_content)
    os.system("chmod a+rwx " + batch_file)
    os.system("chmod a+rwx " + cmd_batch)
    del cmd
    return workdir + "/" + basename + ".batch", logfile


def create_batch_script_slurm(
    cmd,
    workdir,
    basename,
    partition="general",
    nodes=1,
    cpus_per_task=1,
    ntasks_per_node=1,
    mem="4G",
    time="01:00:00",
    account=None,
    dependency=None,
    write_logfile=True,
):
    """
    Create a Slurm-compatible batch script.

    Parameters
    ----------
    cmd : str
        Command to run.
    workdir : str
        Work directory.
    basename : str
        Base name of batch script files. This will be Slurm job name as well (shown in `squeue`, `sacct`, etc.).
    partition : str, optional
        Name of the Slurm partition (queue) to submit to (e.g., "compute", "general").
    nodes : int, optional
        Number of physical nodes to request.
    cpus_per_task : int, optional
        Number of CPUs to allocate for each task (for multi-threaded tasks).
    ntasks_per_node : int, optional
        Number of tasks per node.
    mem : str, optional
        Total memory per node (e.g., "16G", "64000M").
    time : str, optional
        Maximum wall-clock time for the job in HH:MM:SS format.
    account : str, optional
        Slurm account to charge for the job (if required).
    dependency : str, optional
        Slurm job dependency (e.g., "afterok:12345") to delay job until others complete.
    write_logfile : bool, optional
        If True, write job output and error to {workdir}/logs/{basename}.log.
        If False, redirects output to /dev/null.

    Returns
    -------
    str
        Batch script filename.
    str
        Log file path (or /dev/null).
    """
    env_file = os.path.join(workdir, "activate_env.sh")
    # Generate the env file if it doesn't exist
    if not os.path.exists(env_file):
        generate_activate_env(outfile=env_file)
    os.makedirs(workdir, exist_ok=True)
    batch_file = os.path.join(workdir, f"{basename}.sbatch")
    cmd_batch = os.path.join(workdir, f"{basename}_cmd.batch")
    finished_touch_base = os.path.join(workdir, f".Finished_{basename}")
    finished_touch_error = finished_touch_base + "_1"
    finished_touch_success = finished_touch_base + "_0"
    # Remove old status files
    os.system(f"rm -rf {finished_touch_base}*")
    # === Command script content with environment activation ===
    cmd_file_content = "\n".join(
        [
            f"#!/bin/bash",
            f"source {env_file}",
            f"{cmd}",
            "exit_code=$?",
            f"if [ $exit_code -ne 0 ]; then touch {finished_touch_error};",
            f"else touch {finished_touch_success}; fi",
        ]
    )
    # Write log output
    if write_logfile:
        log_dir = os.path.join(workdir, "logs")
        os.makedirs(log_dir, exist_ok=True)
        outputfile = os.path.join(log_dir, f"{basename}.log")
    else:
        outputfile = "/dev/null"
    # === SBATCH directives ===
    sbatch_lines = [
        "#!/bin/bash",
        f"#SBATCH --job-name={basename}",
        f"#SBATCH --partition={partition}",
        f"#SBATCH --nodes={nodes}",
        f"#SBATCH --ntasks-per-node={ntasks_per_node}",
        f"#SBATCH --cpus-per-task={cpus_per_task}",
        f"#SBATCH --mem={mem}",
        f"#SBATCH --time={time}",
    ]
    if account:
        sbatch_lines.append(f"#SBATCH --account={account}")
    if dependency:
        sbatch_lines.append(f"#SBATCH --dependency={dependency}")
    if write_logfile:
        sbatch_lines.append(f"#SBATCH --output={outputfile}")
        sbatch_lines.append(f"#SBATCH --error={outputfile}")
    # === Batch script content ===
    batch_script_content = "\n".join(
        [
            *sbatch_lines,
            "",
            f"bash {cmd_batch}",
            f"rm -rf {batch_file}",
            f"rm -rf {cmd_batch}",
        ]
    )
    # Write files
    for f in [cmd_batch, batch_file]:
        if os.path.exists(f):
            os.remove(f)
    with open(cmd_batch, "w") as f:
        f.write(cmd_file_content)
    with open(batch_file, "w") as f:
        f.write(batch_script_content)
    os.chmod(cmd_batch, 0o777)
    os.chmod(batch_file, 0o777)
    return batch_file, outputfile


def generate_activate_env(outfile="activate_env.sh"):
    """
    Generate a shell script that activates the current Python environment.

    This works for both Conda and virtualenv environments and is safe for use in
    non-interactive shells (e.g., Slurm batch jobs) by explicitly sourcing `conda.sh`.

    If conda is not found in $PATH, it will try loading either `anaconda` or `anaconda3` module.

    Parameters
    ----------
    outfile : str
        Path to the shell script to write (default: ./activate_env.sh).

    Returns
    -------
    str
        Output file name
    """
    outfile = Path(outfile).expanduser().resolve()
    putfile = os.path.abspath(outfile)
    lines = ["#!/bin/bash", ""]

    def module_exists(name):
        """Check if a module exists using 'module avail'."""
        try:
            subprocess.run(
                ["module", "avail", name],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                check=False,
            )
            return True
        except Exception:
            return False

    # Conda-based environment
    if "CONDA_DEFAULT_ENV" in os.environ:
        conda_env = os.environ["CONDA_DEFAULT_ENV"]
        lines.append("# === Activate Conda Environment Safely ===")
        lines.append("if ! command -v conda >/dev/null 2>&1; then")
        if module_exists("anaconda"):
            lines.append("    module load anaconda")
        elif module_exists("anaconda3"):
            lines.append("    module load anaconda3")
        else:
            lines.append("    echo 'No Conda module found (anaconda or anaconda3)'")
            lines.append("    exit 1")
        lines.append("fi")
        lines.append("source $(conda info --base)/etc/profile.d/conda.sh")
        lines.append(f"conda activate {conda_env}")
    # Virtualenv-based environment
    elif "VIRTUAL_ENV" in os.environ:
        venv_path = os.environ["VIRTUAL_ENV"]
        lines.append("# === Activate Virtualenv ===")
        lines.append(f"source {venv_path}/bin/activate")
    else:
        python_path = sys.executable
        lines.append(
            "# === No Conda/Virtualenv Detected — Using current Python directly ==="
        )
        lines.append(f"echo 'No Conda or virtualenv detected; using: {python_path}'")
        lines.append(f"export PATH={os.path.dirname(python_path)}:$PATH")
    # Write file
    with open(outfile, "w") as f:
        f.write("\n".join(lines) + "\n")
    os.chmod(outfile, 0o755)
    print(f"Created activation script at: {outfile}")
    return outfile


def wait_for_dask_workers(client, min_worker=1, timeout=60):
    """
    Wait for dask worker
    
    Parameters
    ----------
    client : dask.client
        Dask client
    min_worker : int, optional
        Minimum number of workers
    timeout : float, optional
        Timeout in seconds
    """
    start = time.time()
    while True:
        if len(client.scheduler_info()["workers"]) >= min_worker:
            break
        if time.time() - start > timeout:
            raise TimeoutError("No Dask workers connected within timeout.")
        time.sleep(2)
        
        
def get_dask_client(
    n_jobs,
    dask_dir,
    cpu_frac=0.8,
    mem_frac=0.8,
    ncpu=-1,
    mem=-1,
    spill_frac=0.6,
    min_mem_per_job=-1,
    min_cpu_per_job=1,
    only_cal=False,
    verbose=True,
):
    """
    Create a Dask client optimized for one-task-per-worker execution,
    where each worker is a separate process that can use multiple threads internally.

    Parameters
    ----------
    n_jobs : int
        Number of MS tasks (ideally = number of MS files)
    dask_dir : str
        Dask temporary directory
    cpu_frac : float, optional
        Fraction of total CPUs to use
    mem_frac : float, optinal;
        Fraction of total memory to use
    ncpu : int, optional
        Number of CPUs to use (if specified, cpu_frac will be ignored)
    mem : float, optional
        Memory in GB to use (if specified, mem_frac will be ignored)
    spill_frac : float, optional
        Spill to disk at this fraction
    min_mem_per_job : float, optional
        Minimum memory per job
    min_cpu_per_job : int, optional
        Minimum CPU threads per job
    only_cal : bool, optional
        Only calculate number of workers
    verbose : bool, optional
        Verbose (details of cluster)

    Returns
    -------
    client : dask.distributed.Client
        Dask clinet
    cluster : dask.distributed.LocalCluster
        Dask cluster
    n_workers : int
        Number of workers
    threads_per_worker : int
        Threads per worker to use
    str
        Dask directory
    """
    logging.getLogger("distributed").setLevel(logging.ERROR)
    # Create the Dask temporary working directory if it does not already exist
    dask_dir=dask_dir.rstrip("/")
    os.makedirs(dask_dir,exist_ok=True)
    dask_dir = tempfile.mkdtemp(prefix=f"{dask_dir}/dask_")
    os.makedirs(dask_dir, exist_ok=True)
    dask_dir_tmp = dask_dir + "/tmp"
    os.makedirs(dask_dir_tmp, exist_ok=True)

    # Detect total system resources
    total_cpus = psutil.cpu_count(logical=True)  # Total logical CPU cores
    total_mem = psutil.virtual_memory().total  # Total system memory (bytes)
    if ncpu >= 1:
        cpu_frac = float(ncpu / total_cpus)
    if mem > 0:
        mem_frac = float((mem * 1024**3) / total_mem)
    if cpu_frac > 0.8:
        if verbose:
            print(
                "Given CPU fraction is more than 80%. Resetting to 80% to avoid system crash."
            )
        cpu_frac = 0.8
    if mem_frac > 0.8:
        if verbose:
            print(
                "Given memory fraction is more than 80%. Resetting to 80% to avoid system crash."
            )
        mem_frac = 0.8

    ############################################
    # Wait until enough free CPU is available
    ############################################
    count = 0
    while True:
        available_cpu_pct = 100 - psutil.cpu_percent(
            interval=1
        )  # Percent CPUs currently free
        available_cpus = int(
            total_cpus * available_cpu_pct / 100.0
        )  # Number of free CPU cores
        usable_cpus = max(
            1, int(total_cpus * cpu_frac)
        )  # Target number of CPU cores we want available based on cpu_frac
        if available_cpus >= int(
            0.5 * usable_cpus
        ):  # Enough free CPUs (at-least more than 50%), exit loop
            usable_cpus = min(usable_cpus, available_cpus)
            break
        else:
            if count == 0:
                print("Waiting for available free CPUs...")
            time.sleep(5)  # Wait a bit and retry
        count += 1

    ############################################
    # Wait until enough free memory is available
    ############################################
    count = 0
    while True:
        available_mem = (
            psutil.virtual_memory().available
        )  # Current available system memory (bytes)
        usable_mem = total_mem * mem_frac  # Target usable memory based on mem_frac
        if (
            available_mem >= 0.5 * usable_mem
        ):  # Enough free memory, (at-least more than 50%) exit loop
            usable_mem = min(usable_mem, available_mem)
            break
        else:
            if count == 0:
                print("Waiting for available free memory...")
            time.sleep(5)  # Wait and retry
        count += 1

    ############################################
    # Calculate memory per worker
    ############################################
    mem_per_worker = usable_mem / n_jobs  # Assume initially one job per worker
    # Apply minimum memory per worker constraint
    min_mem_per_job = round(
        min_mem_per_job, 2
    )  # Ensure min_mem_per_job is a clean float
    if min_mem_per_job > 0 and mem_per_worker < (min_mem_per_job * 1024**3):
        # If calculated memory per worker is smaller than user-requested
        # minimum, adjust number of workers
        if verbose:
            print(
                f"Total memory per job is smaller than {min_mem_per_job} GB. Adjusting total number of workers to meet this."
            )
        mem_per_worker = (
            min_mem_per_job * 1024**3
        )  # Reset memory per worker to minimum allowed
        n_workers = min(
            n_jobs, int(usable_mem / mem_per_worker)
        )  # Reduce number of workers accordingly
    else:
        # Otherwise, just keep n_jobs workers
        n_workers = n_jobs

    #########################################
    # Cap number of workers to available CPUs
    n_workers = max(
        1, min(n_workers, int(usable_cpus / min_cpu_per_job))
    )  # Prevent CPU oversubscription
    # Recalculate final memory per worker based on capped n_workers
    mem_per_worker = usable_mem / n_workers
    # Calculate threads per worker
    threads_per_worker = max(
        1, usable_cpus // max(1, n_workers)
    )  # Each worker gets min_cpu_per_job or more threads

    ##########################################
    if not only_cal and verbose:
        print("#################################")
        print(
            f"Dask workers: {n_workers}, Threads per worker: {threads_per_worker}, Mem/worker: {round(mem_per_worker/(1024.0**3),2)} GB"
        )
        print("#################################")
    # Memory control settings
    swap = psutil.swap_memory()
    swap_gb = swap.total / 1024.0**3
    if swap_gb > 16:
        pass
    elif swap_gb > 4:
        spill_frac = 0.6
    else:
        spill_frac = 0.5

    if spill_frac > 0.7:
        spill_frac = 0.7
    if only_cal:
        return None, None, n_workers, threads_per_worker, round(mem_per_worker/(1024.0**3),2), dask_dir

    soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
    new_soft = min(int(hard * 0.8), hard)  # safe cap
    if soft < new_soft:
        resource.setrlimit(resource.RLIMIT_NOFILE, (new_soft, hard))
    dask.config.set({"temporary-directory": dask_dir})
    cluster = LocalCluster(
        n_workers=n_workers,
        threads_per_worker=1,
        # one python-thread per worker, in workers OpenMP threads can be used
        memory_limit=f"{round(mem_per_worker/(1024.0**3),2)}GB",
        local_directory=dask_dir,
        processes=True,  # one process per worker
        dashboard_address=":0",
        env={
            "TMPDIR": dask_dir_tmp,
            "TMP": dask_dir_tmp,
            "TEMP": dask_dir_tmp,
            "DASK_TEMPORARY_DIRECTORY": dask_dir,
            "MALLOC_TRIM_THRESHOLD_": "0",
        },  # Explicitly set for workers
    )
    client = Client(cluster, timeout="60s", heartbeat_interval="5s")
    dask.config.set(
        {
            "distributed.worker.memory.target": spill_frac,
            "distributed.worker.memory.spill": spill_frac + 0.1,
            "distributed.worker.memory.pause": spill_frac + 0.2,
            "distributed.worker.memory.terminate": spill_frac + 0.25,
        }
    )
    if verbose:
        print(f"Dask dashboard available at: {client.dashboard_link}")

    client.run_on_scheduler(gc.collect)
    return client, cluster, n_workers, threads_per_worker, round(mem_per_worker/(1024.0**3),2), dask_dir


def run_limited_memory_task(task, dask_dir="/tmp", timeout=30):
    """
    Run a task for a limited time, then kill and return memory usage.

    Parameters
    ----------
    task : dask.delayed
        Dask delayed task object
    timeout : int
        Time in seconds to let the task run

    Returns
    -------
    float
        Memory used by task (in GB)
    """
    dask.config.set({"temporary-directory": dask_dir})
    warnings.filterwarnings("ignore")
    cluster = LocalCluster(
        n_workers=1,
        threads_per_worker=1,
        # one python-thread per worker, in workers OpenMP threads can be used
        local_directory=dask_dir,
        processes=True,  # one process per worker
        dashboard_address=":0",
    )
    client = Client(cluster)
    future = client.compute(task)
    start = time.time()

    def get_worker_memory_info():
        proc = psutil.Process()
        return {
            "rss_GB": proc.memory_info().rss / 1024**3,
            "vms_GB": proc.memory_info().vms / 1024**3,
        }

    while not future.done():
        if time.time() - start > timeout:
            try:
                mem_info = client.run(get_worker_memory_info)
                total_rss = sum(v["rss_GB"] for v in mem_info.values())
                per_worker_mem = total_rss
            except Exception as e:
                per_worker_mem = None
            future.cancel()
            client.close()
            cluster.close()
            return per_worker_mem
        time.sleep(1)
    mem_info = client.run(get_worker_memory_info)
    total_rss = sum(v["rss_GB"] for v in mem_info.values())
    per_worker_mem = total_rss
    client.close()
    cluster.close()
    return round(per_worker_mem, 2)


# Exposing only functions
__all__ = [
    name
    for name, obj in globals().items()
    if isinstance(obj, types.FunctionType) and obj.__module__ == __name__
]



import types
import shutil
import traceback
import platform
import ctypes
import tempfile
import os
from contextlib import contextmanager


POSIX_FADV_DONTNEED = 4
libc = ctypes.CDLL("libc.so.6")


#####################################
# Resource management
#####################################
def drop_file_cache(filepath, verbose=False):
    """
    Advise the OS to drop the given file from the page cache.
    Safe, per-file, no sudo required.
    """
    if platform.system() != "Linux":
        raise NotImplementedError("drop_file_cache is only supported on Linux")
    try:
        if not os.path.isfile(filepath):
            return
        fd = os.open(filepath, os.O_RDONLY)
        result = libc.posix_fadvise(fd, 0, 0, POSIX_FADV_DONTNEED)
        os.close(fd)
        if verbose:
            if result == 0:
                print(f"[cache drop] Released: {filepath}")
            else:
                print(f"[cache drop] Failed ({result}) for: {filepath}")
    except Exception as e:
        if verbose:
            print(f"[cache drop] Error for {filepath}: {e}")
            traceback.print_exc()


def drop_cache(path, verbose=False):
    """
    Drop file cache for a file or all files under a directory.

    Parameters
    ----------
    path : str
        File or directory path
    """
    if platform.system() != "Linux":
        raise NotImplementedError("drop_file_cache is only supported on Linux")
    if os.path.isfile(path):
        drop_file_cache(path, verbose=verbose)
    elif os.path.isdir(path):
        for root, _, files in os.walk(path):
            for f in files:
                full_path = os.path.join(root, f)
                drop_file_cache(full_path, verbose=verbose)
    else:
        if verbose:
            print(f"[cache drop] Path does not exist or is not valid: {path}")


def has_space(path, required_gb):
    try:
        stat = shutil.disk_usage(path)
        return (stat.free / 1e9) >= required_gb
    except BaseException:
        return False


@contextmanager
def shm_or_tmp(required_gb, workdir, prefix="solar_", verbose=False):
    """
    Create a temporary working directory:
    1. Try /dev/shm if it has required space
    2. Else TMPDIR if set and has space
    4. Else work directory
    Temporarily sets TMPDIR to the selected path during the context.
    Cleans up after use.

    Parameters
    ----------
    required_gb : float
        Required disk space in GB
    workdir : str
        Fall back work directory
    prefix : str, optional
        Temp directory prefix
    verbose : bool, optional
        Verbose
    """
    candidates = []
    if has_space("/dev/shm", required_gb):
        candidates.append("/dev/shm")
    tmpdir_env = os.environ.get("TMPDIR")
    if tmpdir_env is not None and has_space(tmpdir_env, required_gb):
        candidates.append(tmpdir_env)
    candidates.append(os.getcwd())
    for i in range(len(candidates)):
        base_dir = candidates[i]
        try:
            temp_dir = tempfile.mkdtemp(dir=base_dir, prefix=prefix)
            if verbose:
                if i == 0:
                    print("Using RAM")
                elif i == 1:
                    print("Using {os.environ.get('TMPDIR')}")
                else:
                    print("Using {os.getcwd()}")
            break
        except Exception as e:
            print(f"[shm_or_tmp] Failed to create temp dir in {base_dir}: {e}")
    else:
        raise RuntimeError(
            "Could not create a temporary directory in any fallback location."
        )
    # Override TMPDIR
    old_tmpdir = os.environ.get("TMPDIR")
    os.environ["TMPDIR"] = temp_dir
    try:
        yield temp_dir
    finally:
        # Restore TMPDIR
        if old_tmpdir is not None:
            os.environ["TMPDIR"] = old_tmpdir
        else:
            os.environ.pop("TMPDIR", None)
        # Clean up the temp directory
        try:
            shutil.rmtree(temp_dir)
        except Exception as e:
            print(f"[cleanup] Warning: could not delete {temp_dir}: {e}")


@contextmanager
def tmp_with_cache_rel(required_gb, workdir, prefix="solar_", verbose=False):
    """
    Combined context manager:
    - Uses shm_or_tmp() for workspace
    - Drops kernel page cache for all files in that directory on exit

    Parameters
    ----------
    required_gb : float
        Required disk space in GB
    workdir : str
        Fall back work directory
    prefix : str, optional
        Temp directory prefix
    verbose : bool, optional
        Verbose

    """
    with shm_or_tmp(required_gb, workdir, prefix=prefix, verbose=verbose) as tempdir:
        try:
            yield tempdir
        finally:
            if platform.system() == "Linux":
                drop_cache(tempdir)


def limit_threads(n_threads=-1):
    """
    Limit number of threads usuage

    Parameters
    ----------
    n_threads : int, optional
        Number of threads
    """
    if n_threads > 0:
        os.environ["OMP_NUM_THREADS"] = str(n_threads)
        os.environ["OPENBLAS_NUM_THREADS"] = str(n_threads)
        os.environ["MKL_NUM_THREADS"] = str(n_threads)
        os.environ["VECLIB_MAXIMUM_THREADS"] = str(n_threads)


# Exposing only functions
__all__ = [
    name
    for name, obj in globals().items()
    if isinstance(obj, types.FunctionType) and obj.__module__ == __name__
]



import types
import psutil
import numpy as np
import glob
import os
from casatools import msmetadata, ms as casamstool, table, measures
from .basic_utils import *
from .resource_utils import *

##########################
# Measurement set metadata
##########################


def get_phasecenter(msname, field):
    """
    Get phasecenter of the measurement set

    Parameters
    ----------
    msname : str
        Name of the measurement set
    field : str
        Field name

    Returns
    -------
    float
        RA in degree
    float
        DEC in degree
    """
    msmd = msmetadata()
    msmd.open(msname)
    phasecenter = msmd.phasecenter()
    msmd.close()
    msmd.done()
    radeg = np.rad2deg(phasecenter["m0"]["value"])
    radeg = radeg % 360
    decdeg = np.rad2deg(phasecenter["m1"]["value"])
    decdeg = decdeg % 360
    return round(radeg, 5), round(decdeg, 5)


def get_timeranges_for_scan(
    msname, scan, time_interval, time_window, quack_timestamps=-1
):
    """
    Get time ranges for a scan with certain time intervals

    Parameters
    ----------
    msname : str
        Name of the measurement set
    scan : int
        Scan number
    time_interval : float
        Time interval in seconds
    time_window : float
        Time window in seconds
    quack_timestamps : int, optional
        Number of timestamps ignored at the start and end of each scan

    Returns
    -------
    list
        List of time ranges
    """
    msmd = msmetadata()
    msmd.open(msname)
    try:
        times = msmd.timesforscan(int(scan))
    except BaseException:
        times = msmd.timesforspws(0)
    msmd.close()
    msmd.done()
    time_ranges = []
    if quack_timestamps > 0:
        times = times[quack_timestamps:-quack_timestamps]
    else:
        times = times[1:-1]
    start_time = times[0]
    end_time = times[-1]
    if time_interval < 0 or time_window < 0:
        t = (
            mjdsec_to_timestamp(start_time, str_format=1)
            + "~"
            + mjdsec_to_timestamp(end_time, str_format=1)
        )
        time_ranges.append(t)
        return time_ranges
    total_time = end_time - start_time
    timeres = total_time / len(times)
    ntime_chunk = int(total_time / time_interval)
    ntime = int(time_window / timeres)
    start_time = times[:-ntime]
    indices = np.linspace(0, len(start_time) - 1, num=ntime_chunk, dtype=int)
    timelist = []
    for i in indices:
        try:
            timelist.append(start_time[i])
        except:
            pass
    if len(timelist) == 0:
        time_range = [f"{mjdsec_to_timestamp(times[0], str_format=1)}"]
        return time_range
    for t in timelist:
        time_ranges.append(
            f"{mjdsec_to_timestamp(t, str_format=1)}~{mjdsec_to_timestamp(t+time_window, str_format=1)}"
        )
    return time_ranges


def calc_fractional_bandwidth(msname):
    """
    Calculate fractional bandwidh

    Parameters
    ----------
    msname : str
        Name of measurement set

    Returns
    -------
    float
        Fraction bandwidth in percentage
    """
    msmd = msmetadata()
    msmd.open(msname)
    freqs = msmd.chanfreqs(0)
    bw = max(freqs) - min(freqs)
    frac_bandwidth = bw / msmd.meanfreq(0)
    msmd.close()
    return round(frac_bandwidth * 100.0, 2)


def baseline_names(msname):
    """
    Get baseline names

    Parameters
    ----------
    msname : str
        Measurement set name

    Returns
    -------
    list
        Baseline names list
    """
    mstool = casamstool()
    mstool.open(msname)
    ants = mstool.getdata(["antenna1", "antenna2"])
    mstool.close()
    baseline_ids = set(zip(ants["antenna1"], ants["antenna2"]))
    baseline_names = []
    for ant1, ant2 in sorted(baseline_ids):
        baseline_names.append(str(ant1) + "&&" + str(ant2))
    return baseline_names


def get_ms_size(msname, only_autocorr=False):
    """
    Get measurement set total size on-disk
    (Note: it could be smaller than actual data size, because of data compression)

    Parameters
    ----------
    msname : str
        Measurement set name
    only_autocorr : bool, optional
        Only auto-correlation

    Returns
    -------
    float
        Size in GB
    """
    total_size = 0
    for dirpath, dirnames, filenames in os.walk(msname):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            total_size += os.path.getsize(fp)
    if only_autocorr:
        msmd = msmetadata()
        msmd.open(msname)
        nant = msmd.nantennas()
        msmd.close()
        all_baselines = (nant * nant) / 2
        total_size /= all_baselines
        total_size *= nant
    return round(total_size / (1024**3), 2)  # in GB


def get_column_size(msname, only_autocorr=False):
    """
    Get datacolumn size
    (Note: this is true datasize in memory)

    Parameters
    ----------
    msname : str
        Measurement set
    only_autocorr : bool, optional
        Only auto-correlations

    Returns
    -------
    float
        A single datacolumn data size in GB
    """
    msmd = msmetadata()
    msmd.open(msname)
    nrow = int(msmd.nrows())
    nchan = msmd.nchan(0)
    npol = msmd.ncorrforpol()[0]
    nant = msmd.nantennas()
    msmd.close()
    datasize = nrow * nchan * npol * 16 / (1024.0**3)
    if only_autocorr:
        all_baselines = (nant * nant) / 2
        datasize /= all_baselines
        datasize *= nant
    return round(datasize, 2)


def get_ms_scan_size(msname, scan, only_autocorr=False):
    """
    Get measurement set scan size

    Parameters
    ----------
    msname : str
        Measurement set
    scan : int
        Scan number
    only_autocorr : bool, optional
        Only for auto-correlations

    Returns
    -------
    float
        Size in GB
    """
    tb = table()
    tb.open(msname)
    nrow = tb.nrows()
    tb.close()
    mstool = casamstool()
    mstool.open(msname)
    mstool.select({"scan_number": int(scan)})
    scan_nrow = mstool.nrow(True)
    mstool.close()
    ms_size = get_column_size(msname, only_autocorr=only_autocorr)
    scan_size = scan_nrow * (ms_size / nrow)
    return round(scan_size, 2)


def get_chunk_size(msname, memory_limit=-1, only_autocorr=False):
    """
    Get time chunk size for a memory limit

    Parameters
    ----------
    msname : str
        Measurement set
    memory_limit : int, optional
        Memory limit
    only_autocorr : bool, optional
        Only aut-correlation

    Returns
    -------
    int
        Number of chunks
    """
    if memory_limit == -1:
        memory_limit = psutil.virtual_memory().available / 1024**3  # In GB
    col_size = get_column_size(msname, only_autocorr=only_autocorr)
    nchunk = int(col_size / memory_limit)
    if nchunk < 1:
        nchunk = 1
    return nchunk


def check_datacolumn_valid(msname, datacolumn="DATA"):
    """
    Check whether a data column exists and valid

    Parameters
    ----------
    msname : str
        Measurement set
    datacolumn : str, optional
        Data column string in table (e.g.,DATA, CORRECTED_DATA', MODEL_DATA, FLAG, WEIGHT, WEIGHT_SPECTRUM, SIGMA, SIGMA_SPECTRUM)

    Returns
    -------
    bool
        Whether valid data column is present or not
    """
    tb = table()
    msname = msname.rstrip("/")
    msname = os.path.abspath(msname)
    try:
        tb.open(msname)
        colnames = tb.colnames()
        if datacolumn not in colnames:
            tb.close()
            return False
        try:
            model_data = tb.getcol(datacolumn, startrow=0, nrow=1)
            tb.close()
            if model_data is None or model_data.size == 0:
                return False
            elif (model_data == 0).all():
                return False
            else:
                return True
        except BaseException:
            tb.close()
            return False
    except BaseException:
        return False


def get_bad_ants(msname="", fieldnames=[], n_threads=-1, dry_run=False):
    """
    Get bad antennas

    Parameters
    ----------
    msname : str
        Name of the ms
    fieldnames : list, optional
        Fluxcal field names

    Returns
    -------
    list
        Bad antenna list
    str
        Bad antenna string
    """
    limit_threads(n_threads=n_threads)
    from casatasks import visstat

    if len(fieldnames) == 0:
        print("Provide field names.")
        return [], ""
    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    msname = msname.rstrip("/")
    mspath = os.path.dirname(os.path.abspath(msname))
    os.chdir(mspath)
    msmd = msmetadata()
    all_field_bad_ants = []
    msmd.open(msname)
    nant = msmd.nantennas()
    msmd.close()
    msmd.done()
    for field in fieldnames:
        ant_medians = []
        bad_ants = []
        for ant in range(nant):
            stat_median = visstat(
                vis=msname,
                field=str(field),
                uvrange="0lambda",
                antenna=str(ant) + "&&" + str(ant),
                useflags=False,
            )["DATA_DESC_ID=0"]["median"]
            ant_medians.append(stat_median)
        ant_medians = np.array(ant_medians)
        all_ant_median = np.nanmean(ant_medians)
        all_ant_std = np.nanstd(ant_medians)
        pos = np.where(ant_medians < all_ant_median - (5 * all_ant_std))[0]
        if len(pos) > 0:
            for b_ant in pos:
                bad_ants.append(b_ant)
        all_field_bad_ants.append(bad_ants)
    bad_ants = [set(sublist) for sublist in all_field_bad_ants]
    common_elements = set.intersection(*bad_ants)
    bad_ants = list(common_elements)
    if len(bad_ants) > 0:
        bad_ants_str = ",".join([str(i) for i in bad_ants])
    else:
        bad_ants_str = ""
    return bad_ants, bad_ants_str


def get_common_spw(spw1, spw2):
    """
    Return common spectral windows in merged CASA string format.

    Parameters
    ----------
    spw1 : str
        First spectral window (0:xx~yy)
    spw2 : str
        Second spectral window (0:xx1~yy1)

    Returns
    -------
    str
        Merged spectral window
    """
    from itertools import groupby
    from collections import defaultdict

    def to_set(s):
        out, cur = set(), None
        for part in s.split(";"):
            if ":" in part:
                cur, rng = part.split(":")
            else:
                rng = part
            cur = int(cur)
            a, *b = map(int, rng.split("~"))
            out.update((cur, i) for i in range(a, (b[0] if b else a) + 1))
        return out

    def to_str(pairs):
        spw_dict = defaultdict(list)
        for spw, ch in sorted(pairs):
            spw_dict[spw].append(ch)
        result = []
        for spw, chans in spw_dict.items():
            chans.sort()
            for _, g in groupby(enumerate(chans), lambda x: x[1] - x[0]):
                grp = list(g)
                a, b = grp[0][1], grp[-1][1]
                result.append(f"{a}" if a == b else f"{a}~{b}")
        if len(result) > 0:
            return "0:" + ";".join(result)
        else:
            return ""

    return to_str(to_set(spw1) & to_set(spw2))


def scans_in_timerange(msname="", timerange="", dry_run=False):
    """
    Get scans in the given timerange

    Parameters
    ----------
    msname : str
        Measurement set
    timerange : str
        Time range with date and time

    Returns
    -------
    dict
        Scan dict for timerange
    """
    from casatools import ms, quanta

    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    msname = msname.rstrip("/")
    mspath = os.path.dirname(os.path.abspath(msname))
    os.chdir(mspath)
    qa = quanta()
    ms_tool = ms()
    ms_tool.open(msname)
    # Get scan summary
    scan_summary = ms_tool.getscansummary()
    # Convert input timerange to MJD seconds
    timerange_list = timerange.split(",")
    valid_scans = {}
    for timerange in timerange_list:
        tr_start_str, tr_end_str = timerange.split("~")
        # Try parsing as date string
        tr_start = timestamp_to_mjdsec(tr_start_str)
        tr_end = timestamp_to_mjdsec(tr_end_str)
        for scan_id, scan_info in scan_summary.items():
            t0_str = scan_info["0"]["BeginTime"]
            t1_str = scan_info["0"]["EndTime"]
            scan_start = qa.convert(qa.quantity(t0_str, "d"), "s")["value"]
            scan_end = qa.convert(qa.quantity(t1_str, "d"), "s")["value"]
            # Check overlap
            if scan_end >= tr_start and scan_start <= tr_end:
                if tr_end >= scan_end:
                    e = scan_end
                else:
                    e = tr_end
                if tr_start <= scan_start:
                    s = scan_start
                else:
                    s = tr_start
                if scan_id in valid_scans.keys():
                    old_t = valid_scans[scan_id].split("~")
                    old_s = timestamp_to_mjdsec(old_t[0])
                    old_e = timestamp_to_mjdsec(old_t[-1])
                    if s > old_s:
                        s = old_s
                    if e < old_e:
                        e = old_e
                valid_scans[int(scan_id)] = (
                    mjdsec_to_timestamp(s, str_format=1)
                    + "~"
                    + mjdsec_to_timestamp(e, str_format=1)
                )
    ms_tool.close()
    return valid_scans


def get_refant(msname="", field="", n_threads=-1, dry_run=False):
    """
    Get reference antenna

    Parameters
    ----------
    msname : str
        Name of the measurement set
    field : str, optional
        Field name

    Returns
    -------
    str
        Reference antenna
    """
    limit_threads(n_threads=n_threads)
    from casatasks import visstat, casalog

    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    msname = msname.rstrip("/")
    mspath = os.path.dirname(os.path.abspath(msname))
    os.chdir(mspath)
    casalog.filter("SEVERE")
    msmd = msmetadata()
    msmd.open(msname)
    nant = msmd.nantennas()
    msmd.close()
    msmd.done()
    antamp = []
    antrms = []
    selected_nant = min(10, int(0.1 * nant))
    selected_nant = min(selected_nant, nant)
    for ant in range(selected_nant):
        ant = str(ant)
        t = visstat(
            vis=msname,
            field=field,
            antenna=ant,
            timeaverage=True,
            timebin="500min",
            timespan="state,scan",
            reportingaxes="field",
        )
        item = str(list(t.keys())[0])
        amp = float(t[item]["median"])
        rms = float(t[item]["rms"])
        antamp.append(amp)
        antrms.append(rms)
    antamp = np.array(antamp)
    antrms = np.array(antrms)
    medamp = np.median(antamp)
    medrms = np.median(antrms)
    goodrms = []
    goodamp = []
    goodant = []
    for i in range(len(antamp)):
        if antamp[i] > medamp:
            goodant.append(i)
            goodamp.append(antamp[i])
            goodrms.append(antrms[i])
    goodrms = np.array(goodrms)
    referenceant = np.argmin(goodrms)
    return str(referenceant)


def get_ms_scans(msname):
    """
    Get scans of the measurement set

    Parameters
    ----------
    msname : str
        Measurement set

    Returns
    -------
    list
        Scan list
    """
    msmd = msmetadata()
    msmd.open(msname)
    scans = msmd.scannumbers().tolist()
    msmd.close()
    return scans


def get_submsname_scans(msname):
    """
    Get sub-MS names for each scans of an multi-MS

    Parameters
    ----------
    msname : str
        Name of the measurement set

    Returns
    -------
    list
        msname list
    list
        scan list
    """
    from casatasks import listpartition

    if os.path.exists(msname + "/SUBMSS") == False:
        print("Input measurement set is not a multi-MS")
        return
    partitionlist = listpartition(vis=msname, createdict=True)
    scans = []
    mslist = []
    for i in range(len(partitionlist)):
        subms = partitionlist[i]
        subms_name = msname + "/SUBMSS/" + subms["MS"]
        mslist.append(subms_name)
        os.system(f"rm -rf {subms_name}/.flagversions")
        scan_number = list(subms["scanId"].keys())[0]
        scans.append(scan_number)
    return mslist, scans


def get_observatory_name(msname):
    """
    Get observatory name

    Parameters
    ----------
    msname : str
        Measurement set

    Returns
    -------
    str
        Observatory name in all upper case
    """
    observatory = ""
    try:
        msmd = msmetadata()
        msmd.open(msname)
        observatory = msmd.observatorynames()[0].upper()
        msmd.close()
    except Exception:
        pass
    return observatory


def get_observatory_coord(msname):
    """
    Get observatory coordinate

    Parameters
    ----------
    msname : str
        Measurement set

    Returns
    -------
    float
        Latitude in degrees
    float
        Longitude in degrees
    float
        Height in meters
    """
    msmd = msmetadata()
    msmd.open(msname)
    position = msmd.observatoryposition()
    me = measures()
    obs_pos = me.observatory(msmd.observatorynames()[0])
    lon = obs_pos["m0"]["value"] * (180.0 / 3.141592653589793)
    lat = obs_pos["m1"]["value"] * (180.0 / 3.141592653589793)
    height = obs_pos["m2"]["value"]
    msmd.close()
    return round(lat, 3), round(lon, 3), round(height, 3)


def get_pol_names(msname, fullpol=True):
    """
    Get correlation names

    Parameters
    ----------
    msname : str
        Measurement set
    fullpol : bool, optional
        Full polarization products or not

    Returns
    -------
    list
        List of cross correlation product names
    """
    CASA_POL_PRODUCTS = {
        1: "I",
        2: "Q",
        3: "U",
        4: "V",
        5: "RR",
        6: "RL",
        7: "LR",
        8: "LL",
        9: "XX",
        10: "XY",
        11: "YX",
        12: "YY",
    }
    msmd = msmetadata()
    msmd.open(msname)
    pols = msmd.corrtypesforpol(0)
    msmd.close()
    pol_names = []
    for p in pols:
        pol_name = CASA_POL_PRODUCTS[int(p)]
        if fullpol is True:
            pol_names.append(pol_name)
        else:
            if pol_name in ["XX", "YY", "RR", "LL", "I"]:
                pol_names.append(pol_name)
            else:
                pass
    return pol_names


# Expose functions and classes
__all__ = [
    name
    for name, obj in globals().items()
    if (
        (isinstance(obj, types.FunctionType) or isinstance(obj, type))
        and obj.__module__ == __name__
    )
]



import types
import julian
import resource
import numpy as np
import os
import astropy.units as u
from astropy.time import Time
from astropy.coordinates import Angle
from datetime import datetime as dt
from contextlib import contextmanager

##########################
# Basic utility funactions
##########################
@contextmanager
def suppress_casa_output():
    """
    Supress CASA terminal output
    """
    with open(os.devnull, "w") as fnull:
        old_stdout = os.dup(1)
        old_stderr = os.dup(2)
        os.dup2(fnull.fileno(), 1)
        os.dup2(fnull.fileno(), 2)
        try:
            yield
        finally:
            os.dup2(old_stdout, 1)
            os.dup2(old_stderr, 2)


def get_cachedir():
    """
    Get cache directory
    """
    homedir = os.environ.get("HOME")
    if homedir is None:
        homedir = os.path.expanduser("~")
    username = os.getlogin()
    cachedir = f"{homedir}/.solarpipe"
    os.makedirs(cachedir, exist_ok=True)
    os.makedirs(f"{cachedir}/pids", exist_ok=True)
    return cachedir


def create_datadir(datadir=""):
    """
    Create data directory

    Parameters
    ----------
    datadir : str, optional
        User provided custom data directory
    """
    cachedir = get_cachedir()
    if datadir == "":
        datadir = f"{cachedir}/solarpipe_data"
    os.makedirs(datadir, exist_ok=True)
    with open(f"{cachedir}/solarpipe_data_dir.txt", "w") as f:
        f.write(str(datadir) + "\n")
    return


def get_datadir():
    """
    Get package data directory

    Returns
    -------
    str
        Data directory
    """
    cachedir = get_cachedir()
    if os.path.exists(f"{cachedir}/solarpipe_data_dir.txt") == False:
        return None
    with open(f"{cachedir}/solarpipe_data_dir.txt", "r") as f:
        datadir = f.read().strip()
    os.makedirs(datadir, exist_ok=True)
    return datadir


def split_into_chunks(lst, target_chunk_size):
    """
    Split a list into equal number of elements

    Parameters
    ----------
    lst : list
        List of numbers
    target_chunk_size: int
        Number of elements per chunk

    Returns
    -------
    list
        Chunked list
    """
    n = len(lst)
    num_chunks = max(1, round(n / target_chunk_size))
    avg_chunk_size = n // num_chunks
    remainder = n % num_chunks

    chunks = []
    start = 0
    for i in range(num_chunks):
        extra = 1 if i < remainder else 0  # Distribute remainder
        end = start + avg_chunk_size + extra
        chunks.append(lst[start:end])
        start = end
    return chunks


def average_timestamp(timestamps):
    """
    Compute the average timestamp using astropy from a list of ISO 8601 strings.

    Parameters
    ----------
    timestamps : list
        timestamps (list of str): List of timestamp strings in 'YYYY-MM-DDTHH:MM:SS' format.

    Returns
    --------
    str
        Average timestamp in 'YYYY-MM-DDTHH:MM:SS' format.
    """
    if len(timestamps) == 0:
        return ""
    times = Time(timestamps, format="isot", scale="utc")
    avg_time = Time(np.mean(times.jd), format="jd", scale="utc")
    return avg_time.isot.split(".")[0]  # Strip milliseconds for clean output


def ceil_to_multiple(n, base):
    """
    Round up to the next multiple

    Parameters
    ----------
    n : float
        The number
    base : float
        Whose multiple will be

    Returns
    -------
    float
        The modified number
    """
    return ((n // base) + 1) * base


def angular_separation_equatorial(ra1, dec1, ra2, dec2):
    """
    Calculate angular seperation between two equatorial coordinates

    Parameters
    ----------
    ra1 : float
        RA of the first coordinate in degree
    dec1 : float
        DEC of the first coordinate in degree
    ra2 : float
        RA of the second coordinate in degree
    dec2 : float
        DEC of the second coordinate in degree

    Returns
    -------
    float
        Angular distance in degree
    """
    # Convert RA and Dec from degrees to radians
    ra1 = np.radians(ra1)
    ra2 = np.radians(ra2)
    dec1 = np.radians(dec1)
    dec2 = np.radians(dec2)
    # Apply the spherical distance formula using NumPy functions
    cos_theta = np.sin(dec1) * np.sin(dec2) + np.cos(dec1) * np.cos(dec2) * np.cos(
        ra1 - ra2
    )
    # Calculate the angular separation in radians
    theta_rad = np.arccos(cos_theta)
    # Convert the angular separation from radians to degrees
    theta_deg = np.degrees(theta_rad)
    return round(theta_deg, 2)


def ra_dec_to_deg(ra_hms, dec_dms):
    """
    Convert RA and Dec from hms and dms format to degrees

    Parameters
    ----------
    ra_hms: str
        Right Ascension in 'hms' format
    dec_dms : str
        Declination in 'dms' format

    Returns
    -------
    tuple
        RA and Dec in degrees
    """
    ra = Angle(ra_hms, unit=u.hourangle)
    dec = Angle(dec_dms, unit=u.deg)
    return round(ra.deg, 4), round(dec.deg, 4)


def ra_dec_to_hms_dms(ra_deg, dec_deg):
    """
    Convert RA and Dec in degrees to hms and dms format

    Parameters
    ----------
    ra_deg : float
        Right Ascension in degrees.
    dec_deg : float
        Declination in degrees.

    Returns
    -------
    tuple
        RA in h:m:s format, Dec in d:m:s format (e.g., '1h5m0s', '1d5m0s').
    """
    # Convert RA to h:m:s
    if ra_deg < 0:
        ra_deg += 360
    ra = Angle(ra_deg, unit=u.deg)
    ra_hms = ra.to_string(unit=u.hourangle, sep=":").split(":")
    ra_hms = ra_hms[0] + "h" + ra_hms[1] + "m" + ra_hms[2] + "s"
    # Convert Dec to d:m:s
    dec = Angle(dec_deg, unit=u.deg)
    dec_dms = dec.to_string(unit=u.deg, sep=":").split(":")
    dec_dms = dec_dms[0] + "d" + dec_dms[1] + "m" + dec_dms[2] + "s"
    return ra_hms, dec_dms


def timestamp_to_mjdsec(timestamp, date_format=0):
    """
    Convert timestamp to mjd second.


    Parameters
    ----------
    timestamp : str
        Time stamp to convert
    date_format : int, optional
        Datetime string format
            0: 'YYYY/MM/DD/hh:mm:ss'

            1: 'YYYY-MM-DDThh:mm:ss'

            2: 'YYYY-MM-DD hh:mm:ss'

            3: 'YYYY_MM_DD_hh_mm_ss'

    Returns
    -------
    float
        Return correspondong MJD second of the day
    """
    if date_format == 0:
        try:
            timestamp_datetime = dt.strptime(timestamp, "%Y/%m/%d/%H:%M:%S.%f")
        except BaseException:
            timestamp_datetime = dt.strptime(timestamp, "%Y/%m/%d/%H:%M:%S")
    elif date_format == 1:
        try:
            timestamp_datetime = dt.strptime(timestamp, "%Y-%m-%dT%H:%M:%S.%f")
        except BaseException:
            timestamp_datetime = dt.strptime(timestamp, "%Y-%m-%dT%H:%M:%S")
    elif date_format == 2:
        try:
            timestamp_datetime = dt.strptime(timestamp, "%Y-%m-%d %H:%M:%S.%f")
        except BaseException:
            timestamp_datetime = dt.strptime(timestamp, "%Y-%m-%d %H:%M:%S")
    elif date_format == 3:
        try:
            timestamp_datetime = dt.strptime(timestamp, "%Y_%m_%d_%H_%M_%S.%f")
        except BaseException:
            timestamp_datetime = dt.strptime(timestamp, "%Y_%m_%d_%H_%M_%S")
    else:
        print("No proper format of timestamp.\n")
        return
    mjd = float(
        "{: .2f}".format(
            (julian.to_jd(timestamp_datetime) - 2400000.5) * (24.0 * 3600.0)
        )
    )
    return mjd


def mjdsec_to_timestamp(mjdsec, str_format=0):
    """
    Convert CASA MJD seceonds to CASA timestamp

    Parameters
    ----------
    mjdsec : float
            CASA MJD seconds
    str_format : int
        Time stamp format (0: yyyy-mm-ddTHH:MM:SS.ff, 1: yyyy/mm/dd/HH:MM:SS.ff, 2: yyyy-mm-dd HH:MM:SS)

    Returns
    -------
    str
            CASA time stamp in UTC at ISOT format
    """
    from casatools import measures, quanta

    me = measures()
    qa = quanta()
    today = me.epoch("utc", "today")
    mjd = np.array(mjdsec) / 86400.0
    today["m0"]["value"] = mjd
    hhmmss = qa.time(today["m0"], prec=8)[0]
    date = qa.splitdate(today["m0"])
    qa.done()
    if str_format == 0:
        utcstring = "%s-%02d-%02dT%s" % (
            date["year"],
            date["month"],
            date["monthday"],
            hhmmss,
        )
    elif str_format == 1:
        utcstring = "%s/%02d/%02d/%s" % (
            date["year"],
            date["month"],
            date["monthday"],
            hhmmss,
        )
    else:
        utcstring = "%s-%02d-%02d %s" % (
            date["year"],
            date["month"],
            date["monthday"],
            hhmmss,
        )
    return utcstring


# Exposing only functions
__all__ = [
    name
    for name, obj in globals().items()
    if isinstance(obj, types.FunctionType) and obj.__module__ == __name__
]



import logging

log = logging.getLogger(__name__)


def select_field_id(field_datasets, field=None):
    """
    Parameters
    ----------
    field_datasets : list of :class:`daskms.Dataset`

    field : None or str
        if None, will try and return the only field in the MS.

        .. code-block:: python

            select_field_id(field_ds, "PKS-1934")

    Returns
    -------
    fields_id : int
        Selected field ID
    """
    if field is None:
        if len(field_datasets) == 1:
            return 0

        names = [
            "%s or %d" % (ds.NAME.values[0], i) for i, ds in enumerate(field_datasets)
        ]

        raise ValueError(
            "No field was provided "
            "but multiple fields are present "
            "in the Measurement Set. "
            "Please select a field %s." % names
        )
    elif not isinstance(field, str):
        raise TypeError("field must be None or str")

    for i, ds in enumerate(field_datasets):
        names = ds.NAME.values
        assert len(names) == 1, "Should have a single name per row"

        if str(names[0]) == field or str(i) == field:
            return i

    names = ["%s or %d" % (ds.NAME.values[0], i) for i, ds in enumerate(field_datasets)]

    raise ValueError(
        "%s was requested, but no matching field " "was found %s" % (field, names)
    )


def filter_datasets(datasets, field_id):
    """
    Parameters
    ----------
    datasets : list of :class:`daskms.Dataset`
        List of datasets representing partitions of the Measurement Set.
        Should be partitioned by FIELD_ID and DATA_DESC_ID at least
    field_id : list of ints
        List of valid fields

    Returns
    -------
    datasets : list of :class:`daskms.Dataset`
        Filtered list of datasets
    """
    return [ds for ds in datasets if ds.FIELD_ID == field_id]



import types
import psutil
import traceback
import tempfile
import time
import glob
import os
from .basic_utils import *

####################
# uDOCKER related
####################


def set_udocker_env():
    datadir = get_datadir()
    udocker_dir = datadir + "/udocker"
    os.makedirs(udocker_dir, exist_ok=True)
    os.environ["UDOCKER_DIR"] = udocker_dir
    os.environ["UDOCKER_TARBALL"] = datadir + "/udocker-englib-1.2.11.tar.gz"


def init_udocker():
    set_udocker_env()
    os.system("udocker install")


def check_udocker_container(name):
    """
    Check whether a docker container is present or not

    Parameters
    ----------
    name : str
        Container name

    Returns
    -------
    bool
        Whether present or not
    """
    set_udocker_env()
    pid = os.getpid()
    timestamp = int(time.time() * 1000)
    tmp1 = f"tmp1_{pid}_{timestamp}.txt"
    tmp2 = f"tmp2_{pid}_{timestamp}.txt"
    b = os.system(
        f"udocker --insecure --quiet inspect " + name + f" >> {tmp1} >> {tmp2}"
    )
    os.system(f"rm -rf {tmp1} {tmp2}")
    if b != 0:
        return False
    else:
        return True


def initialize_wsclean_container(name="solarwsclean"):
    """
    Initialize WSClean container

    Parameters
    ----------
    name : str
        Name of the container

    Returns
    -------
    bool
        Whether initialized successfully or not
    """
    set_udocker_env()
    image_name = "devojyoti96/wsclean-solar:latest"
    check_cmd = f"udocker images | grep -q '{image_name}'"
    image_exists = os.system(check_cmd) == 0
    if not image_exists:
        a = os.system(f"udocker pull {image_name}")
    else:
        print(f"Image '{image_name}' already present.")
        a = 0
    if a == 0:
        a = os.system(f"udocker create --name={name} {image_name}")
        print(f"Container started with name : {name}")
        return name
    else:
        print(f"Container could not be created with name : {name}")
        return


def run_wsclean(
    wsclean_cmd,
    container_name="solarwsclean",
    check_container=False,
    verbose=False,
    dry_run=False,
):
    """
    Run WSClean inside a udocker container (no root permission required).

    Parameters
    ----------
    wsclean_cmd : str
        Full WSClean command as a string.
    container_name : str, optional
        Container name
    check_container : bool, optional
        Check container presence or not
    verbose : bool, optional
        Verbose output or not

    Returns
    -------
    int
        Success message
    """
    set_udocker_env()
    pid = os.getpid()
    timestamp = int(time.time() * 1000)
    tmp1 = f"tmp1_{pid}_{timestamp}.txt"
    tmp2 = f"tmp2_{pid}_{timestamp}.txt"

    def show_file(path):
        try:
            print(open(path).read())
        except Exception as e:
            print(f"Error: {e}")

    if check_container:
        container_present = check_udocker_container(container_name)
        if not container_present:
            container_name = initialize_wsclean_container(name=container_name)
            if container_name is None:
                print(
                    f"Container {container_name} is not initiated. First initiate container and then run."
                )
                return 1

    if dry_run:
        cmd = f"chgenter >> {tmp1} >> {tmp2}"
        cwd = os.getcwd()
        full_command = (
            f"udocker --quiet run --nobanner --volume={cwd}:{cwd} solarwsclean {cmd}"
        )
        os.system(full_command)
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        os.system(f"rm -rf {tmp1} {tmp2}")
        return mem

    msname = wsclean_cmd.split(" ")[-1]
    msname = os.path.abspath(msname)
    mspath = os.path.dirname(msname)
    temp_docker_path = tempfile.mkdtemp(prefix="wsclean_udocker_", dir=mspath)
    wsclean_cmd_args = wsclean_cmd.split(" ")[:-1]
    if "-fits-mask" in wsclean_cmd_args:
        index = wsclean_cmd_args.index("-fits-mask")
        name = wsclean_cmd_args[index + 1]
        namedir = os.path.dirname(os.path.abspath(name))
        basename = os.path.basename(os.path.abspath(name))
        wsclean_cmd_args.remove(name)
        wsclean_cmd_args.insert(index + 1, temp_docker_path + "/" + basename)
    if "-name" not in wsclean_cmd_args:
        wsclean_cmd_args.append(
            "-name " + temp_docker_path + "/" + os.path.basename(msname).split(".ms")[0]
        )
    else:
        index = wsclean_cmd_args.index("-name")
        name = wsclean_cmd_args[index + 1]
        namedir = os.path.dirname(os.path.abspath(name))
        basename = os.path.basename(os.path.abspath(name))
        wsclean_cmd_args.remove(name)
        wsclean_cmd_args.insert(index + 1, temp_docker_path + "/" + basename)
    if "-temp-dir" not in wsclean_cmd_args:
        wsclean_cmd_args.append("-temp-dir " + temp_docker_path)
    else:
        index = wsclean_cmd_args.index("-temp-dir")
        name = os.path.abspath(wsclean_cmd_args[index + 1])
        wsclean_cmd_args.remove(name)
        wsclean_cmd_args.insert(index + 1, temp_docker_path)
    wsclean_cmd = (
        " ".join(wsclean_cmd_args)
        + " "
        + temp_docker_path
        + "/"
        + os.path.basename(msname)
    )
    try:
        full_command = f"udocker run --nobanner --volume={mspath}:{temp_docker_path} --workdir {temp_docker_path} solarwsclean {wsclean_cmd}"
        if not verbose:
            full_command += f" >> {mspath}/{tmp1} "
        else:
            print(wsclean_cmd + "\n")
        exit_code = os.system(full_command)
        if exit_code != 0:
            print("##########################")
            print(os.path.basename(msname))
            print("##########################")
            show_file(f"{mspath}/{tmp1}")
        os.system(f"rm -rf {temp_docker_path} {mspath}/{tmp1}")
        return 0 if exit_code == 0 else 1
    except Exception as e:
        os.system(f"rm -rf {temp_docker_path}")
        traceback.print_exc()
        return 1


def run_solar_sidereal_cor(
    msname="",
    only_uvw=False,
    container_name="solarwsclean",
    check_container=False,
    verbose=False,
    dry_run=False,
):
    """
    Run chgcenter inside a udocker container to correct solar sidereal motion (no root permission required).

    Parameters
    ----------
    msname : str
        Name of the measurement set
    only_uvw : bool, optional
        Update only UVW values
        Note: This is required when visibilities are properly phase rotated in correlator to track the Sun,
        but while creating the MS, UVW values are estimated using the first phasecenter of the Sun.
    check_container : bool, optional
        Check container
    container_name : str, optional
        Container name
    verbose : bool, optional
        Verbose output or not

    Returns
    -------
    int
        Success message
    """
    set_udocker_env()
    pid = os.getpid()
    timestamp = int(time.time() * 1000)
    tmp1 = f"tmp1_{pid}_{timestamp}.txt"
    tmp2 = f"tmp2_{pid}_{timestamp}.txt"
    if check_container:
        container_present = check_udocker_container(container_name)
        if not container_present:
            container_name = initialize_wsclean_container(name=container_name)
            if container_name is None:
                print(
                    f"Container {container_name} is not initiated. First initiate container and then run."
                )
                return 1

    if dry_run:
        cmd = f"chgcentre >> {tmp1} >> {tmp2}"
        cwd = os.getcwd()
        full_command = (
            f"udocker --quiet run --nobanner --volume={cwd}:{cwd} solarwsclean {cmd}"
        )
        os.system(full_command)
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        os.system(f"rm -rf {tmp1} {tmp2}")
        return mem

    msname = os.path.abspath(msname)
    mspath = os.path.dirname(msname)
    temp_docker_path = tempfile.mkdtemp(prefix="chgcenter_udocker_", dir=mspath)
    if only_uvw:
        cmd = (
            "chgcentre -only-uvw -solarcenter "
            + temp_docker_path
            + "/"
            + os.path.basename(msname)
        )
    else:
        cmd = (
            "chgcentre -solarcenter "
            + temp_docker_path
            + "/"
            + os.path.basename(msname)
        )
    try:
        full_command = f"udocker --quiet run --nobanner --volume={mspath}:{temp_docker_path} --workdir {temp_docker_path} solarwsclean {cmd}"
        if not verbose:
            full_command += f" >> {tmp1} >> {tmp2}"
        else:
            print(cmd)
        with suppress_casa_output():
            exit_code = os.system(full_command)
        os.system(f"rm -rf {temp_docker_path} {tmp1} {tmp2}")
        return 0 if exit_code == 0 else 1
    except Exception as e:
        os.system(f"rm -rf {temp_docker_path} {tmp1} {tmp2}")
        traceback.print_exc()
        return 1


def run_chgcenter(
    msname,
    ra,
    dec,
    only_uvw=False,
    container_name="solarwsclean",
    verbose=False,
    dry_run=False,
):
    """
    Run chgcenter inside a udocker container (no root permission required).

    Parameters
    ----------
    msname : str
        Name of the measurement set
    ra : str
        RA can either be 00h00m00.0s or 00:00:00.0
    dec : str
        Dec can either be 00d00m00.0s or 00.00.00.0
    only_uvw : bool, optional
        Update only UVW values
        Note: This is required when visibilities are properly phase rotated in correlator,
        but while creating the MS, UVW values are estimated using a wrong phase center.
    container_name : str, optional
        Container name
    verbose : bool, optional
        Verbose output

    Returns
    -------
    int
        Success message
    """
    set_udocker_env()
    pid = os.getpid()
    timestamp = int(time.time() * 1000)
    tmp1 = f"tmp1_{pid}_{timestamp}.txt"
    tmp2 = f"tmp2_{pid}_{timestamp}.txt"
    container_present = check_udocker_container(container_name)
    if not container_present:
        container_name = initialize_wsclean_container(name=container_name)
        if container_name is None:
            print(
                f"Container {container_name} is not initiated. First initiate container and then run."
            )
            return 1
    if dry_run:
        cmd = f"chgenter >> {tmp1} >> {tmp2}"
        cwd = os.getcwd()
        full_command = (
            f"udocker --quiet run --nobanner --volume={cwd}:{cwd} solarwsclean {cmd}"
        )
        os.system(full_command)
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        os.system(f"rm -rf {tmp1} {tmp2}")
        return mem
    msname = os.path.abspath(msname)
    mspath = os.path.dirname(msname)
    temp_docker_path = tempfile.mkdtemp(prefix="chgcenter_udocker_", dir=mspath)
    if only_uvw:
        cmd = (
            "chgcentre -only-uvw "
            + temp_docker_path
            + "/"
            + os.path.basename(msname)
            + " "
            + ra
            + " "
            + dec
        )
    else:
        cmd = (
            "chgcentre "
            + temp_docker_path
            + "/"
            + os.path.basename(msname)
            + " "
            + ra
            + " "
            + dec
        )
    try:
        full_command = f"udocker --quiet run --nobanner --volume={mspath}:{temp_docker_path} --workdir {temp_docker_path} solarwsclean {cmd}"
        if not verbose:
            full_command += f" >> {tmp1} >> {tmp2}"
        else:
            print(cmd)
        exit_code = os.system(full_command)
        os.system(f"rm -rf {temp_docker_path} {tmp1} {tmp2}")
        return 0 if exit_code == 0 else 1
    except Exception as e:
        os.system(f"rm -rf {temp_docker_path} {tmp1} {tmp2}")
        traceback.print_exc()
        return 1
    return


# Expose functions and classes
__all__ = [
    name
    for name, obj in globals().items()
    if (
        (isinstance(obj, types.FunctionType) or isinstance(obj, type))
        and obj.__module__ == __name__
    )
]



import types
import psutil
import numpy as np
import glob
import os
import traceback
from casatools import msmetadata, ms as casamstool, table
from .basic_utils import *
from .resource_utils import *

#############################
# General CASA tasks
#############################


def check_scan_in_caltable(caltable, scan):
    """
    Check scan number available in caltable or not

    Parameters
    ----------
    caltable : str
        Name of the caltable
    scan : int
        Scan number

    Returns
    -------
    bool
        Whether scan is present in the caltable or not
    """
    tb = table()
    tb.open(caltable)
    scans = tb.getcol("SCAN_NUMBER")
    tb.close()
    if int(scan) in scans:
        return True
    else:
        return False


def reset_weights_and_flags(
    msname="", restore_flag=True, force_reset=False, n_threads=-1, dry_run=False
):
    """
    Reset weights and flags for the ms

    Parameters
    ----------
    msname : str
        Measurement set
    restore_flag : bool, optional
        Restore flags or not
    force_reset : bool, optional
        Force reset
    """
    limit_threads(n_threads=n_threads)
    from casatasks import flagdata

    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    msname = msname.rstrip("/")
    if os.path.exists(f"{msname}/.reset") == False or force_reset:
        mspath = os.path.dirname(os.path.abspath(msname))
        os.chdir(mspath)
        if restore_flag:
            print(f"Restoring flags of measurement set : {msname}")
            if os.path.exists(msname + ".flagversions"):
                os.system("rm -rf " + msname + ".flagversions")
            flagdata(vis=msname, mode="unflag", flagbackup=False)
        print(f"Resetting previous weights of the measurement set: {msname}")
        msmd = msmetadata()
        msmd.open(msname)
        npol = msmd.ncorrforpol()[0]
        msmd.close()
        tb = table()
        tb.open(msname, nomodify=False)
        colnames = tb.colnames()
        nrows = tb.nrows()
        if "WEIGHT" in colnames:
            print(f"Resetting weight column to ones of measurement set : {msname}.")
            weight = np.ones((npol, nrows))
            tb.putcol("WEIGHT", weight)
        if "SIGMA" in colnames:
            print(f"Resetting sigma column to ones of measurement set: {msname}.")
            sigma = np.ones((npol, nrows))
            tb.putcol("SIGMA", sigma)
        if "WEIGHT_SPECTRUM" in colnames:
            print(f"Removing weight spectrum of measurement set: {msname}.")
            tb.removecols("WEIGHT_SPECTRUM")
        if "SIGMA_SPECTRUM" in colnames:
            print(f"Removing sigma spectrum of measurement set: {msname}.")
            tb.removecols("SIGMA_SPECTRUM")
        tb.flush()
        tb.close()
        os.system(f"touch {msname}/.reset")
    return


def correct_missing_col_subms(msname):
    """
    Correct for missing colurmns in sub-MSs

    Parameters
    ----------
    msname : str
        Name of the measurement set
    """
    tb = table()
    colname_list = []
    sub_mslist = glob.glob(msname + "/SUBMSS/*.ms")
    for ms in sub_mslist:
        tb.open(ms)
        colname_list.append(tb.colnames())
        tb.close()
    sets = [set(sublist) for sublist in colname_list]
    if len(sets) > 0:
        common_elements = set.intersection(*sets)
        unique_elements = set.union(*sets) - common_elements
        for ms in sub_mslist:
            tb.open(ms, nomodify=False)
            colnames = tb.colnames()
            for colname in unique_elements:
                if colname in colnames:
                    print(f"Removing column: {colname} from sub-MS: {ms}")
                    tb.removecols(colname)
            tb.flush()
            tb.close()
    return


def single_mstransform(
    msname="",
    outputms="",
    field="",
    scan="",
    width=1,
    timebin="",
    datacolumn="DATA",
    spw="",
    corr="",
    timerange="",
    numsubms="auto",
    n_threads=-1,
    dry_run=False,
):
    """
    Perform mstransform of a single scan

    Parameters
    ----------
    msname : str
        Name of the measurement set
    outputms : str
        Output ms name
    scan : int
        Scan to split (a single scan)
    field : str, optional
        Field name
    width : int, optional
        Number of channels to average
    timebin : str, optional
        Time to average
    datacolumn : str, optional
        Data column to split
    spw : str, optional
        Spectral window
    corr : str, optional
        Correlation to split
    timerange : str, optional
        Time range
    numsubms : str, optional
        Number of subms
    n_threads : int, optional
        Number of CPU threads

    Returns
    -------
    str
        Output measurement set name
    """
    limit_threads(n_threads=n_threads)
    from casatasks import mstransform, initweights, flagdata

    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    if timebin == "" or timebin is None:
        timeaverage = False
    else:
        timeaverage = True
    if width > 1:
        chanaverage = True
    else:
        chanaverage = False
    outputms = outputms.rstrip("/")
    if os.path.exists(outputms):
        os.system("rm -rf " + outputms)
    if os.path.exists(outputms + ".flagversions"):
        os.system("rm -rf " + outputms + ".flagversions")
    try:
        if n_threads < 1:
            n_threads = 2
        else:
            n_threads = min(n_threads, 2)
        if field == "":
            msmd = msmetadata()
            msmd.open(msname)
            field = str(msmd.fieldsforscan(int(scan))[0])
            msmd.close()
        with suppress_casa_output():
            mstransform(
                vis=msname,
                outputvis=outputms,
                spw=spw,
                timerange=timerange,
                field=field,
                scan=scan,
                datacolumn=datacolumn,
                createmms=True,
                correlation=corr,
                timeaverage=timeaverage,
                timebin=timebin,
                chanaverage=chanaverage,
                chanbin=int(width),
                nthreads=n_threads,
                separationaxis="scan",
                numsubms=numsubms,
            )
        with suppress_casa_output():
            initweights(vis=outputms, wtmode="ones", dowtsp=True)
            flagdata(
                vis=outputms,
                mode="clip",
                clipzeros=True,
                datacolumn="data",
                flagbackup=False,
            )
        os.system(f"touch {outputms}/.splited")
        return outputms
    except Exception as e:
        traceback.print_exc()
        if os.path.exists(outputms):
            os.system("rm -rf " + outputms)
        return


# Expose functions and classes
__all__ = [
    name
    for name, obj in globals().items()
    if (
        (isinstance(obj, types.FunctionType) or isinstance(obj, type))
        and obj.__module__ == __name__
    )
]



from .basic_utils import *
from .calibration import *
from .casatasks import *
from .flagging import *
from .image_utils import *
from .imaging import *
from .logger_utils import *
from .meer_ploting_utils import *
from .meer_utils import *
from .ms_metadata import *
from .proc_manage_utils import *
from .resource_utils import *
from .selfcal_utils import *
from .sunpos_utils import *
from .udocker_utils import *
from .prefect_logger_utils import *
from .prefect_setup_utils import *




import logging
import resource
import psutil
import dask
import numpy as np
import argparse
import traceback
import time
import sys
import os
import copy
from casatasks import casalog
from casatools import msmetadata, table
from dask import delayed
from dask.distributed import Client
from functools import partial
from meersolar.utils import *

logging.getLogger("distributed").setLevel(logging.WARNING)

try:
    casalogfile = casalog.logfile()
    os.system("rm -rf " + casalogfile)
except BaseException:
    pass

datadir = get_datadir()


def do_selfcal(
    msname="",
    workdir="",
    selfcaldir="",
    start_threshold=5,
    end_threshold=3,
    max_iter=100,
    max_DR=1000,
    min_iter=2,
    DR_convegerence_frac=0.3,
    uvrange="",
    minuv=0,
    solint="60s",
    weight="briggs",
    robust=0.0,
    do_apcal=True,
    min_tol_factor=1.0,
    applymode="calonly",
    solar_selfcal=True,
    ncpu=-1,
    mem=-1,
    dry_run=False,
    logfile="selfcal.log",
):
    """
    Do selfcal iterations and use convergence rules to stop

    Parameters
    ----------
    msname : str
        Name of the measurement set
    workdir : str
        Work directory
    selfcaldir : str
        Working directory
    start_threshold : int, optional
        Start CLEAN threhold
    end_threshold : int, optional
        End CLEAN threshold
    max_iter : int, optional
        Maximum numbers of selfcal iterations
    max_DR : float, optional
        Maximum dynamic range
    min_iter : int, optional
        Minimum numbers of seflcal iterations at different stages
    DR_convegerence_frac : float, optional
        Dynamic range fractional change to consider as converged
    uvrange : str, optional
        UV-range for calibration
    minuv : float, optionial
        Minimum UV-lambda to use in imaging
    solint : str, optional
        Solutions interval
    weight : str, optional
        Imaging weighting
    robust : float, optional
        Briggs weighting robust parameter (-1 to 1)
    do_apcal : bool, optional
        Perform ap-selfcal or not
    min_tol_factor : float, optional
         Minimum tolerable variation in temporal direction in percentage
    applymode : str, optional
        Solution apply mode
    solar_selfcal : bool, optional
        Whether is is solar selfcal or not
    ncpu : int, optional
        Number of CPU threads to use
    mem : float, optional
        Memory in GB to use
    logfile : str, optional
        Log file name

    Returns
    -------
    int
        Success message
    str
        Final caltable
    """
    limit_threads(n_threads=ncpu)
    from casatasks import split, flagdata, flagmanager

    if dry_run:
        process = psutil.Process(os.getpid())
        mem = round(process.memory_info().rss / 1024**3, 2)  # in GB
        return mem
    sub_observer = None
    logger, logfile = create_logger(
        os.path.basename(logfile).split(".log")[0], logfile, verbose=False
    )
    if os.path.exists(f"{workdir}/jobname_password.npy") and logfile is not None:
        time.sleep(5)
        jobname, password = np.load(
            f"{workdir}/jobname_password.npy", allow_pickle=True
        )
        if os.path.exists(logfile):
            sub_observer = init_logger(
                "remotelogger_selfcal_{os.path.basename(msname).split('.ms')[0]}",
                logfile,
                jobname=jobname,
                password=password,
            )
    try:
        msname = os.path.abspath(msname.rstrip("/"))
        selfcaldir = selfcaldir.rstrip("/")
        os.makedirs(selfcaldir, exist_ok=True)

        os.chdir(selfcaldir)
        selfcalms = selfcaldir + "/selfcal_" + os.path.basename(msname)
        if os.path.exists(selfcalms):
            os.system("rm -rf " + selfcalms)
        if os.path.exists(selfcalms + ".flagversions"):
            os.system("rm -rf " + selfcalms + ".flagversions")

        ##############################
        # Restoring any previous flags
        ##############################
        with suppress_casa_output():
            flags = flagmanager(vis=msname, mode="list")
        keys = flags.keys()
        for k in keys:
            if k == "MS":
                pass
            else:
                version = flags[0]["name"]
                try:
                    with suppress_casa_output():
                        flagmanager(vis=msname, mode="restore", versionname=version)
                        flagmanager(vis=msname, mode="delete", versionname=version)
                except BaseException:
                    pass
        if os.path.exists(msname + ".flagversions"):
            os.system("rm -rf " + msname + ".flagversions")

        ##############################
        # Spliting corrected data
        ##############################
        hascor = check_datacolumn_valid(msname, datacolumn="CORRECTED_DATA")
        msmd = msmetadata()
        msmd.open(msname)
        scan = int(msmd.scannumbers()[0])
        field = int(msmd.fieldsforscan(scan)[0])
        freqMHz = msmd.meanfreq(0, unit="MHz")
        msmd.close()
        if hascor:
            logger.info(f"Spliting corrected data to ms : {selfcalms}")
            with suppress_casa_output():
                split(
                    vis=msname,
                    field=str(field),
                    scan=str(scan),
                    outputvis=selfcalms,
                    datacolumn="corrected",
                )
        else:
            logger.info(f"Spliting data to ms : {selfcalms}")
            with suppress_casa_output():
                split(
                    vis=msname,
                    field=str(field),
                    scan=str(scan),
                    outputvis=selfcalms,
                    datacolumn="data",
                )
        msname = selfcalms

        ##########################################
        # Initiate proper weighting
        ##########################################
        with suppress_casa_output():
            flagdata(
                vis=msname,
                mode="clip",
                clipzeros=True,
                datacolumn="data",
                flagbackup=False,
            )

        ############################################
        # Imaging and calibration parameters
        ############################################
        logger.info(f"Estimating imaging Parameters ...")
        cellsize = calc_cellsize(msname, 3)
        instrument_fov = calc_field_of_view(msname, FWHM=False)
        sun_size = calc_sun_dia(freqMHz)
        fov = min(
            instrument_fov, 2 * sun_size * 60
        )  # 2 times sun size at that frequency
        imsize = int(fov / cellsize)
        pow2 = np.ceil(np.log2(imsize)).astype("int")
        possible_sizes = []
        for p in range(pow2):
            for k in [3, 5]:
                possible_sizes.append(k * 2**p)
        possible_sizes = np.sort(np.array(possible_sizes))
        possible_sizes = possible_sizes[possible_sizes >= imsize]
        imsize = max(1024, int(possible_sizes[0]))
        unflagged_antenna_names, flag_frac_list = get_unflagged_antennas(msname)
        refant = unflagged_antenna_names[0]

        ############################################
        # Initiating selfcal Parameters
        ############################################
        logger.info(f"Estimating self-calibration parameters...")
        DR1 = 0.0
        DR2 = 0.0
        DR3 = 0.0
        RMS1 = -1.0
        RMS2 = -1.0
        RMS3 = -1.0
        num_iter = 0
        num_iter_after_ap = 0
        num_iter_fixed_sigma = 0
        last_sigma_DR1 = 0
        sigma_reduced_count = 0
        calmode = "p"
        threshold = start_threshold
        last_round_gaintable = ""
        use_previous_model = False
        os.system("rm -rf *_selfcal_present*")
        ###########################################
        # Starting selfcal loops
        ##########################################
        while True:
            ##################################
            # Selfcal round parameters
            ##################################
            logger.info("######################################")
            logger.info(
                f"Selfcal iteration : "
                + str(num_iter)
                + ", Threshold: "
                + str(threshold)
                + ", Calibration mode: "
                + str(calmode)
            )
            msg, gaintable, dyn, rms, final_image, final_model, final_residual = (
                intensity_selfcal(
                    msname,
                    logger,
                    selfcaldir,
                    cellsize,
                    imsize,
                    round_number=num_iter,
                    uvrange=uvrange,
                    minuv=minuv,
                    calmode=calmode,
                    solint=solint,
                    refant=str(refant),
                    applymode=applymode,
                    min_tol_factor=min_tol_factor,
                    threshold=threshold,
                    use_previous_model=use_previous_model,
                    weight=weight,
                    robust=robust,
                    use_solar_mask=solar_selfcal,
                    ncpu=ncpu,
                    mem=round(mem, 2),
                )
            )
            if msg == 1:
                if num_iter == 0:
                    logger.info(
                        f"No model flux is picked up in first round. Trying with lowest threshold.\n"
                    )
                    (
                        msg,
                        gaintable,
                        dyn,
                        rms,
                        final_image,
                        final_model,
                        final_residual,
                    ) = intensity_selfcal(
                        msname,
                        logger,
                        selfcaldir,
                        cellsize,
                        imsize,
                        round_number=num_iter,
                        uvrange=uvrange,
                        minuv=minuv,
                        calmode=calmode,
                        solint=solint,
                        refant=str(refant),
                        applymode=applymode,
                        min_tol_factor=min_tol_factor,
                        threshold=end_threshold,
                        use_previous_model=False,
                        weight=weight,
                        robust=robust,
                        use_solar_mask=solar_selfcal,
                        ncpu=ncpu,
                        mem=round(mem, 2),
                    )
                    if msg == 1:
                        os.system("rm -rf *_selfcal_present*")
                        time.sleep(5)
                        clean_shutdown(sub_observer)
                        return msg, []
                    else:
                        threshold = end_threshold
                else:
                    os.system("rm -rf *_selfcal_present*")
                    return msg, []
            if msg == 2:
                os.system("rm -rf *_selfcal_present*")
                time.sleep(5)
                clean_shutdown(sub_observer)
                return msg, []
            if num_iter == 0:
                DR1 = DR3 = DR2 = dyn
                RMS1 = RMS2 = RMS3 = rms
            elif num_iter == 1:
                DR3 = dyn
                RMS2 = RMS1
                RMS1 = rms
            else:
                DR1 = DR2
                DR2 = DR3
                DR3 = dyn
                RMS3 = RMS2
                RMS2 = RMS1
                RMS1 = rms
            logger.info(
                f"RMS based dynamic ranges: "
                + str(DR1)
                + ","
                + str(DR2)
                + ","
                + str(DR3)
            )
            logger.info(
                f"RMS of the images: " + str(RMS1) + "," + str(RMS2) + "," + str(RMS3)
            )
            if DR3 > 0.9 * DR2:
                use_previous_model = True
            else:
                use_previous_model = False

            #########################################################
            # If DR is decreasing (DR decrease in phase-only selfcal)
            #########################################################
            if (
                (DR3 < 0.85 * DR2 and DR3 < 0.9 * DR1 and DR2 > DR1)
                and calmode == "p"
                and num_iter > min_iter
            ):
                logger.info(f"Dynamic range decreasing in phase-only self-cal.")
                if do_apcal:
                    logger.info(f"Changed calmode to 'ap'.")
                    calmode = "ap"
                    if threshold > end_threshold and num_iter_fixed_sigma > min_iter:
                        threshold -= 1
                        sigma_reduced_count += 1
                        num_iter_fixed_sigma = 0
                else:
                    os.system("rm -rf *_selfcal_present*")
                    time.sleep(5)
                    clean_shutdown(sub_observer)
                    return 0, last_round_gaintable

            ##############################################################
            # If DR is decreasing (DR decrease in amplitude-phase selfcal)
            ##############################################################
            if (
                (DR3 < 0.9 * DR2 and DR2 > 1.5 * DR1)
                and calmode == "ap"
                and num_iter_after_ap > min_iter
            ):
                logger.info(
                    f"Dynamic range is decreasing after minimum numbers of 'ap' rounds.\n"
                )
                os.system("rm -rf *_selfcal_present*")
                time.sleep(5)
                clean_shutdown(sub_observer)
                return 0, last_round_gaintable

            ###########################
            # If maximum DR has reached
            ###########################
            if DR3 >= max_DR and num_iter_after_ap > min_iter:
                logger.info(f"Maximum dynamic range is reached.\n")
                os.system("rm -rf *_selfcal_present*")
                time.sleep(5)
                clean_shutdown(sub_observer)
                return 0, gaintable

            ##########################
            # If DR suddenly decreased
            ##########################
            if DR3 < 0.7 * DR2 and do_apcal == True and sigma_reduced_count > 1:
                logger.info(
                    f"Dynamic range dropped suddenly. Using last round caltable as final.\n"
                )
                os.system("rm -rf *_selfcal_present*")
                time.sleep(5)
                clean_shutdown(sub_observer)
                return 0, last_round_gaintable
            ###########################
            # Checking DR convergence
            ###########################
            # Condition 1
            # (If DR did not increase after one round of sigma reduction, do not reduce sigma further and exit)
            ###########################
            elif (
                ((do_apcal and calmode == "ap") or do_apcal == False)
                and num_iter_fixed_sigma > min_iter
                and (
                    last_sigma_DR1 > 0
                    and abs(round(np.nanmedian([DR1, DR2, DR3]), 0) - last_sigma_DR1)
                    / last_sigma_DR1
                    < DR_convegerence_frac
                )
                and sigma_reduced_count > 1
            ):
                if threshold > end_threshold:
                    logger.info(
                        f"DR does not increase over last two changes in threshold, but minimum threshold has not reached yet.\n"
                    )
                    logger.info(
                        f"Starting final self-calibration rounds with threshold = "
                        + str(end_threshold)
                        + "sigma...\n"
                    )
                    threshold = end_threshold
                    sigma_reduced_count += 1
                    num_iter_fixed_sigma = 0
                    continue
                else:
                    logger.info(
                        f"Selfcal converged. DR does not increase over last two changes in threshold.\n"
                    )
                    os.system("rm -rf *_selfcal_present*")
                    time.sleep(5)
                    clean_shutdown(sub_observer)
                    return 0, gaintable
            else:
                ########################################
                # Condition 2
                # If DR does not increase a certain percentage
                ########################################
                if (
                    abs(DR1 - DR2) / DR2 < DR_convegerence_frac
                    and num_iter > min_iter
                    and threshold == end_threshold + 1
                ):
                    #####################################
                    # Change from phase only selfcal to amplitude-phase selfcal
                    #####################################
                    if do_apcal and calmode == "p":
                        logger.info(
                            f"Dynamic range converged. Changing calmode to 'ap'.\n"
                        )
                        calmode = "ap"
                        if num_iter_fixed_sigma > min_iter:
                            threshold -= 1
                            sigma_reduced_count += 1
                            num_iter_fixed_sigma = 0
                    ######################################
                    # Converged if already in apcal
                    ######################################
                    elif (
                        do_apcal and num_iter_after_ap > min_iter
                    ) or do_apcal == False:
                        logger.info(f"Self-calibration has converged.\n")
                        os.system("rm -rf *_selfcal_present*")
                        time.sleep(5)
                        clean_shutdown(sub_observer)
                        return 0, gaintable
                ######################################
                # Condition 3
                # Reducing threshold if not converged
                ######################################
                elif (
                    abs(DR1 - DR2) / DR2 < DR_convegerence_frac
                    and threshold > end_threshold
                    and num_iter_fixed_sigma > min_iter
                ):
                    threshold -= 1
                    logger.info(f"Reducing threshold to : " + str(threshold))
                    sigma_reduced_count += 1
                    num_iter_fixed_sigma = 0
                    if last_sigma_DR1 > 0:
                        last_sigma_DR1 = round(np.nanmean([DR1, DR2, DR3]), 0)
                    else:
                        last_sigma_DR1 = round(np.nanmean([DR1, DR2, DR3]), 0)
                #########################################
                # In apcal and maximum iteration has reached
                #########################################
                elif (
                    (do_apcal == False or (do_apcal and calmode == "ap"))
                    and num_iter > min_iter
                    and num_iter == max_iter
                ):
                    logger.info(
                        f"Self-calibration is finished. Maximum iteration is reached.\n"
                    )
                    os.system("rm -rf *_selfcal_present*")
                    time.sleep(5)
                    clean_shutdown(sub_observer)
                    return 0, gaintable
            num_iter += 1
            last_round_gaintable = gaintable
            if calmode == "ap":
                num_iter_after_ap += 1
            num_iter_fixed_sigma += 1
    except Exception as e:
        traceback.print_exc()
        os.system("rm -rf *_selfcal_present*")
        time.sleep(5)
        clean_shutdown(sub_observer)
        return 1, []


def main(
    mslist,
    workdir,
    caldir,
    start_thresh=5,
    stop_thresh=3,
    max_iter=100,
    max_DR=1000,
    min_iter=2,
    conv_frac=0.3,
    solint="60s",
    uvrange="",
    minuv=0,
    weight="briggs",
    robust=0.0,
    applymode="calonly",
    min_tol_factor=1.0,
    do_apcal=True,
    solar_selfcal=True,
    keep_backup=False,
    cpu_frac=0.8,
    mem_frac=0.8,
    jobid=0,
    start_remote_log=False,
    dask_addr=None,
):
    """
    Perform iterative self-calibration on a list of measurement sets.

    Parameters
    ----------
    mslist : str
        Comma-separated list of target measurement sets to be self-calibrated.
    workdir : str
        Path to the working directory for outputs, intermediate files, and logs.
    caldir : str
        Directory containing calibration tables (e.g., from flux or phase calibrators).
    start_thresh : float, optional
        Initial image dynamic range threshold to start self-calibration. Default is 5.
    stop_thresh : float, optional
        Target dynamic range at which to stop iterative self-calibration. Default is 3.
    max_iter : int, optional
        Maximum number of self-calibration iterations. Default is 100.
    max_DR : float, optional
        Maximum dynamic range allowed before halting iterations. Default is 1000.
    min_iter : int, optional
        Minimum number of iterations before checking for convergence. Default is 2.
    conv_frac : float, optional
        Convergence criterion: fractional change in dynamic range below which iteration stops. Default is 0.3.
    solint : str, optional
        Solution interval for gain calibration (e.g., "inf", "60s", "int"). Default is "60s".
    uvrange : str, optional
        UV range to be used for imaging and calibration, in CASA format. Default is "" (all baselines).
    minuv : float, optional
        Minimum baseline length (in wavelengths) to include. Default is 0.
    weight : str, optional
        Weighting scheme for imaging (e.g., "natural", "uniform", "briggs"). Default is "briggs".
    robust : float, optional
        Robustness parameter for Briggs weighting (ignored if not using "briggs"). Default is 0.0.
    applymode : str, optional
        Apply mode for calibration tables ("calonly", "calflag", etc.). Default is "calonly".
    min_tol_factor : float, optional
        Minimum factor for tolerance comparison during convergence checks. Default is 1.0.
    do_apcal : bool, optional
        Whether to apply polarization and bandpass calibration before starting selfcal. Default is True.
    solar_selfcal : bool, optional
        If True, uses solar-specific masking and flux normalization. Default is True.
    keep_backup : bool, optional
        If True, keeps backup MS before applying selfcal solutions. Default is False.
    cpu_frac : float, optional
        Fraction of available CPUs to use per job. Default is 0.8.
    mem_frac : float, optional
        Fraction of available system memory to use per job. Default is 0.8.
    jobid : int, optional
        Identifier for job tracking and logging. Default is 0.
    start_remote_log : bool, optional
        Whether to initiate remote logging via job credentials. Default is False.
    dask_addr : str, optional
        Dask scheduler address

    Returns
    -------
    int
        Success message
    """
    pid = os.getpid()
    cachedir = get_cachedir()
    save_pid(pid, f"{cachedir}/pids/pids_{jobid}.txt")

    mslist = str(mslist).split(",")
    if workdir == "":
        workdir = os.path.dirname(os.path.abspath(mslist[0])) + "/workdir"
    os.makedirs(workdir, exist_ok=True)

    if caldir == "" or not os.path.exists(caldir):
        caldir = f"{workdir}/caltables"
    os.makedirs(caldir, exist_ok=True)

    os.makedirs(workdir + "/logs", exist_ok=True)
    mainlog_file = workdir + "/logs/selfcal_targets.mainlog"
    mainlogger, mainlog_file = create_logger(
        os.path.basename(mainlog_file).split(".mainlog")[0], mainlog_file, verbose=False
    )
    ############
    # Logger
    ############
    observer = None
    if (
        start_remote_log
        and os.path.exists(f"{workdir}/jobname_password.npy")
        and mainlog_file is not None
    ):
        time.sleep(5)
        jobname, password = np.load(
            f"{workdir}/jobname_password.npy", allow_pickle=True
        )
        if os.path.exists(mainlog_file):
            observer = init_logger(
                "all_selfcal", mainlog_file, jobname=jobname, password=password
            )
    if observer == None:
        print("Remote link or jobname is blank. Not transmiting to remote logger.")
    ###########

    ###########################
    # WSClean container
    ###########################
    container_name = "solarwsclean"
    container_present = check_udocker_container(container_name)
    if not container_present:
        container_name = initialize_wsclean_container(name=container_name)
        if container_name is None:
            print(
                f"Container {container_name} is not initiated. First initiate container and then run."
            )
            return 1
    org_mslist = copy.deepcopy(mslist)
    try:
        if len(mslist) == 0:
            mainlogger.info("Please provide at-least one measurement set.")
            msg = 1
        else:
            mem_limit = do_selfcal(dry_run=True)
            partial_do_selfcal = partial(
                do_selfcal,
                start_threshold=float(start_thresh),
                end_threshold=float(stop_thresh),
                max_iter=int(max_iter),
                max_DR=float(max_DR),
                min_iter=int(min_iter),
                DR_convegerence_frac=float(conv_frac),
                uvrange=str(uvrange),
                minuv=float(minuv),
                solint=str(solint),
                weight=str(weight),
                robust=float(robust),
                do_apcal=do_apcal,
                applymode=applymode,
                min_tol_factor=float(min_tol_factor),
                solar_selfcal=solar_selfcal,
            )

            ####################################
            # Filtering any corrupted ms
            #####################################
            filtered_mslist = []  # Filtering in case any ms is corrupted
            for ms in mslist:
                checkcol = check_datacolumn_valid(ms)
                if checkcol:
                    filtered_mslist.append(ms)
                else:
                    mainlogger.info(f"Issue in : {ms}")
                    os.system("rm -rf {ms}")
            mslist = filtered_mslist

            chanlist = []
            for ms in mslist:
                channame = (
                    os.path.basename(ms)
                    .split(".ms")[0]
                    .split("spw_")[-1]
                    .split("_time")[0]
                )
                if channame not in chanlist:
                    chanlist.append(channame)

            available_mem = psutil.virtual_memory().available / 1024**3
            if mem_limit < 4 and available_mem > 4:
                min_mem_per_job = 4
            else:
                min_mem_per_job = mem_limit

            ######################################
            # Resetting maximum file limit
            ######################################
            soft_limit, hard_limit = resource.getrlimit(resource.RLIMIT_NOFILE)
            new_soft_limit = max(soft_limit, int(0.8 * hard_limit))
            if soft_limit < new_soft_limit:
                resource.setrlimit(resource.RLIMIT_NOFILE, (new_soft_limit, hard_limit))

            num_fd_list = []
            if len(mslist) == 0:
                mainlogger.error("No filtered ms to continue.")
                return 1
            else:
                for ms in mslist:
                    msmd = msmetadata()
                    msmd.open(ms)
                    times = msmd.timesforspws(0)
                    timeres = np.diff(times)
                    pos = np.where(timeres > 3 * np.nanmedian(timeres))[0]
                    max_intervals = min(1, len(pos))
                    freqs = msmd.chanfreqs(0, unit="MHz")
                    freqres = np.diff(freqs)
                    pos = np.where(freqres > 3 * np.nanmedian(freqres))[0]
                    max_nchan = min(1, len(pos))
                    msmd.close()
                    per_job_fd = (
                        (max_nchan + 1) * max_intervals * 4 * 2
                    )  # 4 types of images, 2 is fudge factor
                    if per_job_fd == 0:
                        per_job_fd = 1
                    num_fd_list.append(per_job_fd)
                total_fd = max(num_fd_list) * len(mslist)
                n_jobs = max(1, int(new_soft_limit / total_fd))
                n_jobs = min(len(mslist), n_jobs)
                if dask_addr is None:
                    dask_client, dask_cluster, n_jobs, n_threads, mem_limit, dask_dir = (
                        get_dask_client(
                            n_jobs,
                            dask_dir=workdir,
                            cpu_frac=float(cpu_frac),
                            mem_frac=float(mem_frac),
                            min_cpu_per_job=3,
                            min_mem_per_job=min_mem_per_job,
                        )
                    )
                else:
                    _, _, n_jobs, n_threads, mem_limit, dask_dir = (
                        get_dask_client(
                            n_jobs,
                            dask_dir=workdir,
                            cpu_frac=float(cpu_frac),
                            mem_frac=float(mem_frac),
                            min_cpu_per_job=3,
                            min_mem_per_job=min_mem_per_job,
                            only_cal=True,
                        )
                    )
                    dask_client=Client(address=dask_addr)
                    os.system(f"rm -rf {dask_dir}")
                wait_for_dask_workers(dask_client,min_worker=1,timeout=60)
                tasks = []
                for ms in mslist:
                    logfile = (
                        workdir
                        + "/logs/"
                        + os.path.basename(ms).split(".ms")[0]
                        + "_selfcal.log"
                    )
                    mainlogger.info(f"MS name: {ms}, Log file: {logfile}\n")
                    tasks.append(
                        delayed(partial_do_selfcal)(
                            ms,
                            workdir,
                            workdir
                            + "/"
                            + os.path.basename(ms).split(".ms")[0]
                            + "_selfcal",
                            ncpu=n_threads,
                            mem=mem_limit,
                            logfile=logfile,
                        )
                    )
                futures = dask_client.compute(tasks)
                results = list(dask_client.gather(futures))
                dask_client.close()
                if dask_addr is None:
                    dask_cluster.close()
                    os.system(f"rm -rf {dask_dir}")
                gcal_list = []
                for i in range(len(results)):
                    r = results[i]
                    msg = r[0]
                    if msg != 0:
                        mainlogger.info(
                            f"Self-calibration was not successful for ms: {mslist[i]}."
                        )
                    else:
                        gcal = r[1]
                        tb = table()
                        tb.open(gcal)
                        scan = np.unique(tb.getcol("SCAN_NUMBER"))[0]
                        tb.close()
                        final_gain_caltable = caldir + f"/selfcal_scan_{scan}.gcal"
                        os.system(f"cp -r {gcal} {final_gain_caltable}")
                        gcal_list.append(final_gain_caltable)
                if not keep_backup:
                    for ms in mslist:
                        selfcaldir = (
                            workdir
                            + "/"
                            + os.path.basename(ms).split(".ms")[0]
                            + "_selfcal"
                        )
                        os.system("rm -rf " + selfcaldir)
                if len(gcal_list) > 0:
                    mainlogger.info(f"Final selfcal caltables: {gcal_list}")
                    msg = 0
                else:
                    mainlogger.info("No self-calibration is successful.")
                    msg = 1
    except Exception as e:
        traceback.print_exc()
        msg = 1
    finally:
        time.sleep(5)
        for ms in org_mslist:
            drop_cache(ms)
        drop_cache(workdir)
        clean_shutdown(observer)
    return msg


def cli():
    parser = argparse.ArgumentParser(
        description="Self-calibration", formatter_class=SmartDefaultsHelpFormatter
    )

    # Essential parameters
    basic_args = parser.add_argument_group(
        "###################\nEssential parameters\n###################"
    )
    basic_args.add_argument(
        "mslist",
        type=str,
        help="Comma-separated list of measurement sets (required positional argument)",
    )
    basic_args.add_argument(
        "--workdir",
        type=str,
        default="",
        required=True,
        help="Working directory",
    )
    basic_args.add_argument(
        "--caldir",
        type=str,
        default="",
        required=True,
        help="Caltable directory",
    )

    # Advanced parameters
    adv_args = parser.add_argument_group(
        "###################\nAdvanced calibration and imaging parameters\n###################"
    )
    adv_args.add_argument(
        "--start_thresh",
        type=float,
        default=5,
        help="Starting CLEANing threshold",
        metavar="Float",
    )
    adv_args.add_argument(
        "--stop_thresh",
        type=float,
        default=3,
        help="Stop CLEANing threshold",
        metavar="Float",
    )
    adv_args.add_argument(
        "--max_iter",
        type=int,
        default=100,
        help="Maximum number of selfcal iterations",
        metavar="Integer",
    )
    adv_args.add_argument(
        "--max_DR",
        type=float,
        default=1000,
        help="Maximum dynamic range",
        metavar="Float",
    )
    adv_args.add_argument(
        "--min_iter",
        type=int,
        default=2,
        help="Minimum number of selfcal iterations",
        metavar="Integer",
    )
    adv_args.add_argument(
        "--conv_frac",
        type=float,
        default=0.3,
        help="Fractional change in DR to determine convergence",
        metavar="Float",
    )
    adv_args.add_argument("--solint", type=str, default="60s", help="Solution interval")
    adv_args.add_argument(
        "--uvrange",
        type=str,
        default="",
        help="Calibration UV-range (CASA format)",
    )
    adv_args.add_argument(
        "--minuv",
        type=float,
        default=0,
        help="Minimum UV-lambda used for imaging",
        metavar="Float",
    )
    adv_args.add_argument("--weight", type=str, default="briggs", help="Imaging weight")
    adv_args.add_argument(
        "--robust",
        type=float,
        default=0.0,
        help="Robust parameter for briggs weight",
        metavar="Float",
    )
    adv_args.add_argument(
        "--applymode",
        type=str,
        default="calonly",
        help="Solution apply mode",
        metavar="String",
    )
    adv_args.add_argument(
        "--min_tol_factor",
        type=float,
        default=1.0,
        help="Minimum tolerable variation in temporal direction in percentage",
        metavar="Float",
    )
    adv_args.add_argument(
        "--no_apcal",
        action="store_false",
        dest="do_apcal",
        help="Do not perform ap-selfcal",
    )
    adv_args.add_argument(
        "--no_solar_selfcal",
        action="store_false",
        dest="solar_selfcal",
        help="Do not perform solar self-calibration",
    )
    adv_args.add_argument(
        "--keep_backup",
        action="store_true",
        help="Keep backup of self-calibration rounds",
    )
    adv_args.add_argument(
        "--start_remote_log", action="store_true", help="Start remote logging"
    )

    # Resource management parameters
    hard_args = parser.add_argument_group(
        "###################\nHardware resource management parameters\n###################"
    )
    hard_args.add_argument(
        "--cpu_frac",
        type=float,
        default=0.8,
        help="CPU fraction to use",
        metavar="Float",
    )
    hard_args.add_argument(
        "--mem_frac",
        type=float,
        default=0.8,
        help="Memory fraction to use",
        metavar="Float",
    )
    hard_args.add_argument("--jobid", type=int, default=0, help="Job ID")

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    msg = main(
        mslist=args.mslist,
        workdir=args.workdir,
        caldir=args.caldir,
        start_thresh=args.start_thresh,
        stop_thresh=args.stop_thresh,
        max_iter=args.max_iter,
        max_DR=args.max_DR,
        min_iter=args.min_iter,
        conv_frac=args.conv_frac,
        solint=args.solint,
        uvrange=args.uvrange,
        minuv=args.minuv,
        weight=args.weight,
        robust=args.robust,
        applymode=args.applymode,
        min_tol_factor=args.min_tol_factor,
        do_apcal=args.do_apcal,
        solar_selfcal=args.solar_selfcal,
        keep_backup=args.keep_backup,
        cpu_frac=args.cpu_frac,
        mem_frac=args.mem_frac,
        jobid=args.jobid,
        start_remote_log=args.start_remote_log,
    )
    return msg


if __name__ == "__main__":
    result = cli()
    if result > 0:
        result = 1
    print("\n###################\nSelf-calibration is done.\n###################\n")
    os._exit(result)
